﻿/* CorWX - The Core War simulator
 * Copyright (c) 2009 Petr Koupy <petr.koupy@gmail.com>
 * 
 * File: Engine.cs
 * 
 * Description:
 * Engine of MARS (Memory Array Redcode Simulator) compatible with ICWS 94' Draft 3.3.
 * Contains memory loader and main executive function.
 * 
 * Revision history:
 * Rev  Date        Who             Description
 * 1.0  13/08/08    Petr Koupy      Initial draft
 * 1.1  30/04/09    Petr Koupy      WPF adaptation
 * 
 * GNU General Public License Usage
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Windows.Media;
using System.ComponentModel;
using System.Threading;
using System.Text;
using System.IO;

// Address alias for easy modification of core size range.
using Address = System.Int32;

namespace CorWX
{
    /* Graphical user interface is reported of engine events through this
     * engine interface. All communication from within engine is done
     * through interface, thus it is easy to change or switch GUI without 
     * engine modification.
     */
    public interface EngineInterface
    {
        // Cell with specified address is influenced by warrior with specified color.
        void Repaint(Address address, Color color);
        // Specified warrior has added (increment = true) or removed (increment = false) task to/from its task queue.
        void TaskChange(byte warrior, bool increment);
        // Engine is busy from now on.
        void Started();
        // Engine is idle from now on.
        void Finished();
        // All living warriors have executed one instruction.
        void Cycle();
        // Specified error occured within engine.
        void Error(Engine.ErrorState error);
    }

    /* Engine is responsible for executing instructions, loading warrior 
     * into core, task queues management and core management.
     */
    public class Engine
    {
        // Interface property for reporting internal events.
        private static EngineInterface engineInterface;
        public static EngineInterface EngineInterface
        {
            set 
            { 
                Engine.engineInterface = value; 
            }
        }

        /* Engine and GUI are running on separate threads to avoid user
         * interface lags. BackgroundWorker event handler calls engine
         * function to start executing instructions in loop.
         */
        public BackgroundWorker backgroundWorker;
        private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            Start();
        }
        
        /* To determine whether engine should execute only single instruction
         * of each warrior or start executing instructions in loop until the
         * end of the battle or manually stopped.
         */
        public bool step;

        public bool IsBusy; // If set to true, engine is working.

        /* Randomizer is used to generate random instructions during core 
         * wiping and to generate random load addresses of warriors.
         */
        private Random randomizer;

        /* Value represents time in miliseconds for which engine thread
         * would be put into sleep after each execution cycle.
         */
        private Int32 sleepLimit;
        public Int32 SleepLimit
        {
            get
            {
                return sleepLimit;
            }
            set
            {
                if (value >= 0 && value <= Int32.MaxValue)
                {
                    sleepLimit = value;
                }
            }
        }

        /* Return value from the loader and main executive function. Error 
         * value correspond to event which is not compatible with 
         * standard and cause battle halt or impossibility to launch
         * the battle.
         */
        public enum ErrorState
        {
            Normal,
            UndefinedInstruction,
            MinDistance,
            Separation,
            MaxLength
        }

        /* Each warrior is equipped with appropriate color corresponding to
         * its load order. If all colors are used and there is still warriors
         * to load, colors are assigned again from beginning of an array.
         * Colors can be added to array at anytime since there is no presumption
         * about quantity. It is also possible to set a list of colors instead of
         * array and add this functionality to GUI, so user would be able to 
         * manage colors by himself.
         */
        public static Color[] colors = new Color[]
        {
            Colors.Yellow, Colors.Lime,  Colors.Cyan, Colors.Magenta, 
            Colors.Gold, Colors.LightGreen, Colors.DeepSkyBlue, Colors.White,
            Colors.Orange, Colors.SpringGreen, Colors.CornflowerBlue, Colors.HotPink
        };

        /* For further explanation of instruction opcodes see ICWS94 3.3 on line 0736.
         * Static array is used for searching specific opcode during random generation
         * or during determining string identity.
         */
        public static readonly Opcode[] opcode = new Opcode[] 
        { 
          Opcode.DAT, Opcode.MOV, Opcode.ADD, Opcode.SUB, Opcode.MUL,
          Opcode.DIV, Opcode.MOD, Opcode.JMP, Opcode.JMZ, Opcode.JMN,
          Opcode.DJN, Opcode.CMP, Opcode.SNE, Opcode.SLT, Opcode.SPL,
          Opcode.NOP 
        }; 
        public enum Opcode
        {
            DAT, // DATa
            MOV, // MOVe
            ADD, // ADDition
            SUB, // SUBstraction
            MUL, // MULtiplication
            DIV, // DIVision
            MOD, // MODulo
            JMP, // JuMP
            JMZ, // JuMp if Zero
            JMN, // JuMp if Not zero
            DJN, // Decrement and Jump if Not zero
            CMP, // CoMPare alias SEQ - Skip next if EQual
            SNE, // Skip next if Not Equal
            SLT, // Skip next if Lower Than
            SPL, // SPLit
            NOP  // No OPeration
        }

        /* For further explanation of instruction modifiers see ICWS94 3.3 on line 0664.
         * Static array is used for searching specific opcode during random generation
         * or during determining string identity.
         */
        public static readonly Modifier[] modifier = new Modifier[] 
        { 
          Modifier.A, Modifier.B, Modifier.AB, Modifier.BA,
          Modifier.F, Modifier.X, Modifier.I
        }; 
        public enum Modifier
        {
            A,
            B,
            AB,
            BA,
            F,
            X,
            I
        }

        /* For further explanation of instruction modes see ICWS94 3.3 on line 0593.
         * Static array is used for searching specific opcode during random generation
         * or during determining string identity. Order of elements in enumeration 
         * and array must correlate with similar structures in independent parser.
         */
        public static readonly Mode[] mode = new Mode[] 
        { 
          Mode.Immediate, Mode.Direct, Mode.A_indirect, Mode.B_indirect,
          Mode.A_decrement, Mode.B_decrement, Mode.A_increment, Mode.B_increment
        }; 
        public enum Mode
        {
            Immediate,      // #
            Direct,         // $
            A_indirect,     // *
            B_indirect,     // @
            A_decrement,    // {
            B_decrement,    // <
            A_increment,    // }
            B_increment     // >
        }

        // One instruction corresponds to one cell in memory core.
        public struct Instruction
        {
            public Opcode opcode;
            public Modifier modifier;
            public Mode aMode;
            public Address aNumber;
            public Mode bMode;
            public Address bNumber;
        }

        /* Note: Following predefined variables have their set sections
         * not to protect warriors during loading sequence, but to
         * protect engine from crash. Further limitations are calculated
         * in loader and user is notified with error if it is not
         * possible to load warriors with adjusted set of variables.
         */

        // Number of instructions, which can be loaded into core at once.
        private Address coreSize;
        public Address CoreSize
        {
            get
            {
                return coreSize;
            }
            set
            {
                if (value >= 0 && value <= Address.MaxValue)
                {
                    coreSize = value;
                }
            }
        }

        /* How many instructions can be executed from each warrior before
         * declaring a tie.
         */
        private Int32 maxCycles;
        public Int32 MaxCycles
        {
            get
            {
                return maxCycles;
            }
            set
            {
                if (value >= 0 && value <= Int32.MaxValue)
                {
                    maxCycles = value;
                }
            }
        }

        /* Initial instruction is loaded into every cell of memory core
         * prior to loading warriors.
         * Initial instruction modifier stores information for loader. 
         * If value is set to 0, loader operates in normal mode and loads
         * initial instruction into every cell. If value is set to 1, 
         * which means NONE, loader does not reset core cells leaving them 
         * untouched after previous battle. If value is set to 2, which 
         * means RANDOM, loader loads random instruction into each cell. 
         * Note that after program starts, core is initialized with behaviour 
         * of value set to 0.
         */
        public byte InitialInstructionModifier; // 0 (NORMAL), 1 (NONE), 2 (RANDOM)
        private Instruction initialInstruction;
        public Instruction InitialInstruction
        {
            get
            {
                return initialInstruction;
            }
            set
            {
                if (value.aNumber >= 0 && value.aNumber < CoreSize &&
                    value.bNumber >= 0 && value.bNumber < CoreSize)
                {
                    initialInstruction = value;
                }
                else // Operands are out of range. Max possible operands are set.
                {
                    initialInstruction = value;
                    initialInstruction.aNumber = coreSize - 1;
                    initialInstruction.bNumber = coreSize - 1;
                }
            }
        }

        /* Maximum number of instructions allowed per warrior during
         * loading sequence.
         */
        private Address maxLength;
        public Address MaxLength
        {
            get
            {
                return maxLength;
            }
            set
            {
                if (value >= 0 && value <= CoreSize)
                {
                    maxLength = value;
                }
                else // Warrior must not be longer than core size.
                {
                    maxLength = coreSize;
                }
            }
        }

        // Maximum size of warrior's task queue - number of active processes. 
        private Int32 maxProcesses;
        public Int32 MaxProcesses
        {
            get
            {
                return maxProcesses;
            }
            set
            {
                if (value >= 0 && value <= Int32.MaxValue)
                {
                    maxProcesses = value;
                }
            }
        }

        /* The minimum number of instructions from the first instruction
         * of one warrior to the first instruction of the next warrior.
         * This value takes place only if separation modifier is set to
         * 0 (MINDISTANCE). If not then min distance is ignored and separation
         * takes effect.
         */
        private Address minDistance;
        public Address MinDistance
        {
            get
            {
                return minDistance;
            }
            set
            {
                if (value >= 0 && value < CoreSize)
                {
                    minDistance = value;
                }
                else // In order to load at least one warrior, min distance must be lower than core size.
                {
                    minDistance = coreSize - 1;
                }
            }
        }

        /* This is the range available for warriors to read information
         * from core. The range is centered on the current instruction.
         * Attempts to read outside the limits of this range result in 
         * folding target distance by a modulo operation. Although it
         * is not strictly limited in standard, read distance must be a 
         * factor of core size.
         */
        private Address readDistance;
        public Address ReadDistance
        {
            get
            {
                return readDistance;
            }
            set
            {
                if (value >= 0 && value <= CoreSize && CoreSize % value == 0)
                {
                    readDistance = value;
                }
                else // Core size is default value for read distance.
                {
                    readDistance = coreSize;
                }
            }
        }

        /* This is the range available for warriors to write information
         * to core. The range is centered on the current instruction.
         * Attempts to write outside the limits of this range result in 
         * folding target distance by a modulo operation. Although it
         * is not strictly limited in standard, write distance must be 
         * a factor of core size.
         */
        private Address writeDistance;
        public Address WriteDistance
        {
            get
            {
                return writeDistance;
            }
            set
            {
                if (value >= 0 && value <= CoreSize && CoreSize % value == 0)
                {
                    writeDistance = value;
                }
                else // Core size is default value for write distance.
                {
                    writeDistance = coreSize;
                }
            }
        }

        /* The rigid number of instructions from the first instruction
         * of one warrior to the first instruction of the next warrior.
         * This value takes place only if separation modifier is set to
         * 1 (SEPARATION). If not then separation is ignored and min distance
         * takes effect.
         */
        public byte SeparationModifier; // 0 (MINDISTANCE), 1 (SEPARATION)
        private Address separation;
        public Address Separation
        {
            get
            {
                return separation;
            }
            set
            {
                if (value >= 0 && value < CoreSize)
                {
                    separation = value;
                }
                else // In order to load at least one warrior, separation must be lower than core size.
                {
                    separation = coreSize - 1;
                }
            }
        }
        
        // Maximum number of warriors which can be loaded to core at once.
        private byte warriors;
        public byte Warriors
        {
            get
            {
                return warriors;
            }
            set
            {
                if (value >= 0 && value <= byte.MaxValue)
                {
                    warriors = value;
                }
            }
        }

        // Custom variable set loaded by default during engine initialization.
        public void LoadDefault()
        {
            CoreSize = 8000;
            MaxCycles = 8000;
            InitialInstructionModifier = 0;
            initialInstruction.opcode = Opcode.DAT;
            initialInstruction.modifier = Modifier.F;
            initialInstruction.aMode = Mode.Direct;
            initialInstruction.aNumber = 0;
            initialInstruction.bMode = Mode.Direct;
            initialInstruction.bNumber = 0;
            MaxLength = 200;
            MaxProcesses = 400;
            MinDistance = 200;
            ReadDistance = 8000;
            WriteDistance = 8000;
            SeparationModifier = 0;
            Separation = 0;
            Warriors = 12;
            RebuildCore();
        }

        /* Prepares random instructions for loader. Static arrays of 
         * opcodes, modifiers and modes are used to retrieve instruction
         * part corresponding to random value.
         */
        private Instruction PrepareRandomInstruction()
        {
            Instruction instruction;
            instruction.opcode = opcode[randomizer.Next(0, opcode.Length - 1)];
            instruction.modifier = modifier[randomizer.Next(0, modifier.Length - 1)];
            instruction.aMode = mode[randomizer.Next(0, mode.Length - 1)];
            instruction.aNumber = randomizer.Next(0, CoreSize - 1);
            instruction.bMode = mode[randomizer.Next(0, mode.Length - 1)];
            instruction.bNumber = randomizer.Next(0, CoreSize - 1);
            return instruction;
        }

        /* One continuous section of core where is not loaded any warrior and
         * where it is possible to load first instruction of any new warrior
         * in order to avoid violation of min distance or separation limit.
         */
        private struct Gap
        {
            public Address left; // left boundary of gap
            public Address right; // right boundary of gap
            public Gap(Address left, Address right)
            {
                if (left < right)
                {
                    this.left = left;
                    this.right = right;
                }
                else
                {
                    this.left = right;
                    this.right = left;
                }
            }
        }
        /* Represents a map of free space in core. Note that according to gap
         * definition, free space is not mentioned as an address range where 
         * is not loaded any warrior, but an address range where can be loaded
         * first instruction of any new warrior without violating distance limits.
         */
        private List<Gap> gaps;

        /* Main memory core can be accessed from outside readonly by 
         * an auxiliary function.
         */
        private Instruction[] Core;
        public Instruction CoreAt(Address address)
        {
            if (address < Core.Length)
                return (Core[address]);
            else
                return initialInstruction;
        }

        /* Builds up the main core and fills it with initial instructions
         * according to initial instruction modifier. It also resets
         * auxiliary data structures.
         */
        public void RebuildCore()
        {
            Instruction[] oldCore = Core;
            Core = new Instruction[coreSize];
            for (Address i = 0; i < coreSize; i++)
            {
                if (InitialInstructionModifier == 1 && i < oldCore.Length) // NONE
                    Core[i] = oldCore[i]; 
                else if (InitialInstructionModifier == 2) // RANDOM
                    Core[i] = PrepareRandomInstruction();
                else // NORMAL
                    Core[i] = initialInstruction;
            }
            Queues = new List<Warrior>(); // Reset list of warriors.
            gaps = new List<Gap>(); // Reset map of freespace.
            if (SeparationModifier == 0)
                gaps.Add(new Gap(0, coreSize - 1 - minDistance));
            else
                gaps.Add(new Gap(0, coreSize - 1 - separation));
            cycles = 0; // reset cycle counter
        }

        /* Warrior has its own FIFO task queue with appropriate
         * limitation, which is equal to max processes variable in 
         * the time of warrior's creation.
         */
        public class Warrior
        {
            private List<Address> tasks; // FIFO
            private Int32 taskLimit; // FIFO length
            public Color color; // color assigned to warrior by loader
            public byte id; // index of the warrior in warriors list before battle
            public bool selected; // somehow selected/highlighted by GUI

            public Int32 TaskCount
            {
                get
                {
                    return tasks.Count;
                }
            }
            public bool IsAlive()
            {
                if (tasks.Count <= 0)
                    return false; // Is dead.
                else
                    return true; // Is alive.
            }
            public Address PopTask()
            {
                Address task = tasks[0];
                tasks.RemoveAt(0);
                return task;
            }
            public void PushTask(Address task)
            {
                // If queue is full, task is refused.
                if (tasks.Count < taskLimit)
                {
                    tasks.Add(task);
                }
                // Added instruction can be next instruction to execute.
                this.PaintExecute();
            }
            // Repaint instruction which will be executed next.
            public void PaintExecute()
            {
                if (tasks.Count > 0)
                {
                    if (selected == true)
                        engineInterface.Repaint(tasks[0], Colors.Blue); // highlighted
                    else
                        engineInterface.Repaint(tasks[0], Colors.Red); // normal
                }
            }
            public Warrior(Int32 taskLimit, byte id)
            {
                tasks = new List<Address>();
                this.taskLimit = taskLimit;
                this.id = id;
                this.selected = false;
            }
        }

        /* When the warrior is loaded into core, it is added into the 
         * list of warriors. The order in which warriors are added to 
         * list is maintained during execution of their tasks.
         */
        private List<Warrior> Queues;
        /* Since dead warriors are removed from list during execution
         * it is important to access living warriors by id, which is
         * equal to index of warrior in warriors list before battle 
         * starts. If requested warrior is dead or has never existed,
         * return value is set to false.
         */
        public bool QueuesAtID(int id, out Warrior outWarrior)
        {
            foreach (Warrior warrior in Queues)
                if (warrior.id == id)
                {
                    outWarrior = warrior;
                    return true;
                }
            outWarrior = new Warrior(0, 0); // Should not be used if return value is false.
            return false; // ID not found.
        }
        // Represents number of living warriors.
        public Int32 QueuesCount
        {
            get
            {
                return Queues.Count;
            }
        }

        /* Address is folded only if read/write limit is different than
         * size of core. Note that if address is within limit it isn't
         * folded. This fact ensure right behaviour within limit even
         * if read/write limit is not factor of core size. Since in
         * this implementation it is prohibited to set other limits than
         * factors of core size, it is only precaution for future 
         * changes.
         */
        private Address Fold(Address pointer, Address limit)
        {
            Address result;
            if (pointer > coreSize - (limit / 2)) // Pointer is within lower half of limited range.
                return pointer; // not folded
            result = pointer % limit;
            if (result > (limit / 2)) // Pointer is NOT within upper half of limited range.
                result += coreSize - limit; // folded
            return result;
        }

        // Translator between string representatives and corresponding modes.
        private bool whatMode(char identifier, out Mode mode)
        {
            switch (identifier)
            {
                case '#':
                    mode =  Mode.Immediate;
                    return true;
                case '$':
                    mode = Mode.Direct;
                    return true;
                case '*':
                    mode = Mode.A_indirect;
                    return true;
                case '@':
                    mode = Mode.B_indirect;
                    return true;
                case '{':
                    mode = Mode.A_decrement;
                    return true;
                case '<':
                    mode = Mode.B_decrement;
                    return true;
                case '}':
                    mode = Mode.A_increment;
                    return true;
                case '>':
                    mode = Mode.B_increment;
                    return true;
                default:
                    mode = Mode.Direct;
                    return false;
            }
        }

        /* Loader gets load file as a memory stream. While stream is analyzed,
         * each instruction is recognized and saved into array. Last line of
         * load file determines offset to instruction that is added to warrior's
         * task queue as the initial instruction. After warrior is successfully
         * built, load address is calculated according to min distance limit,
         * separation limit, warrior's length and available gaps. Warrior is then
         * loaded into core and added to warriors list.
         */
        public ErrorState Loader(Stream stream)
        {
            List<Instruction> warrior = new List<Instruction>();
            int startInstruction = 0; // offset to initial instruction of warrior
            Address loadAddress = 0;
            StreamReader reader = new StreamReader(stream);
            StringBuilder builder = new StringBuilder();
            
            // Analyzing stream.
            while (!Char.IsDigit((char)reader.Peek()))
            {
                Instruction instruction = new Instruction();
                builder = new StringBuilder();
                
                // opcode
                while ((char)reader.Peek() != '.')
                    builder.Append((char)reader.Read());
                reader.Read();
                for (int i = 0; i < opcode.Length; i++)
                    if (opcode[i].ToString("G") == builder.ToString())
                        instruction.opcode = opcode[i];
                if (instruction.opcode.ToString("G") == "")
                    return ErrorState.UndefinedInstruction;

                // modifier
                builder = new StringBuilder();
                while ((char)reader.Peek() != ' ')
                    builder.Append((char)reader.Read());
                reader.Read();
                for (int i = 0; i < modifier.Length; i++)
                    if (modifier[i].ToString("G") == builder.ToString())
                        instruction.modifier = modifier[i];
                if (instruction.modifier.ToString("G") == "")
                    return ErrorState.UndefinedInstruction;

                // A-mode
                if (!whatMode((char)reader.Read(), out instruction.aMode))
                    return ErrorState.UndefinedInstruction;

                // A-number
                builder = new StringBuilder();
                while ((char)reader.Peek() != ',')
                    builder.Append((char)reader.Read());
                reader.Read();
                if (!Int32.TryParse(builder.ToString(), out instruction.aNumber))
                    return ErrorState.UndefinedInstruction;
                // Illegal value is fixed.
                while (instruction.aNumber < 0)
                    instruction.aNumber += coreSize;
                instruction.aNumber %= coreSize;

                reader.Read(); // ' ' after comma

                // B-mode
                if (!whatMode((char)reader.Read(), out instruction.bMode))
                    return ErrorState.UndefinedInstruction;

                // B-number
                builder = new StringBuilder();
                while ((char)reader.Peek() != '\r')
                    builder.Append((char)reader.Read());
                reader.Read(); // '\r'
                reader.Read(); // '\n'
                if (!Int32.TryParse(builder.ToString(), out instruction.bNumber))
                    return ErrorState.UndefinedInstruction;
                // Illegal value is fixed.
                while (instruction.bNumber < 0)
                    instruction.bNumber += coreSize;
                instruction.bNumber %= coreSize;

                // Instruction successfully recognized.
                if (warrior.Count < maxLength)
                    warrior.Add(instruction);
                else
                    return ErrorState.MaxLength;
            }
            // Offset to initial instruction is stored on last line of load file.
            startInstruction = Int32.Parse(reader.ReadLine());
            if (startInstruction < 0)
                startInstruction = 0; // Illegal value is fixed.

            // Load address evaluation.
            if (SeparationModifier == 1) // rigid separation
            {
                /* According to definition of rigid separation, it is not possible to load
                 * warrior which is longer than separation. Note that if rigid separation
                 * is active, warriors are loaded into core sequentially in clusters
                 * determined by separation limit. To achieve this behaviour, there is only
                 * one gap in list of gaps which represents rest of free space. This gap is
                 * cut from left after successful calculation of load address.
                 */
                if (gaps.Count > 0)
                {
                    Address left = gaps[0].left;
                    Address right = gaps[0].right;
                    if (warrior.Count <= separation + 1 && right - left + 1 >= warrior.Count)
                    {
                        loadAddress = left;
                        if (left + separation + 1 <= right) 
                        {
                            left += separation + 1;
                            gaps.RemoveAt(0);
                            gaps.Add(new Gap(left, right));
                        }
                        else
                            gaps.RemoveAt(0); // Gap is smaller than separation - core is full.
                    }
                    else
                        return ErrorState.Separation; // Warrior is longer than separation or too long for remaining gap.
                }
                else
                    return ErrorState.Separation; // Core is full.
            }
            else // random separation
            {
                /* Each gap is verified and according to the length of warrior is considered 
                 * to be the possible gap for loading. Final gap is then randomly chosen from
                 * possible gaps. Load address is randomly chosen from final gap address range.
                 * Given gap is then cut into two pieces and gaps list is updated.
                 */
                int i = 0;
                Address left = 0;
                Address right = 0;
                List<int> possibleGaps = new List<int>(); // indexes of gaps in gaps list that meet criteria
                // Determining possible gaps.
                while (i < gaps.Count)
                {
                    left = gaps[i].left;
                    right = gaps[i].right;
                    /* If warrior is longer than min distance, right boundary must be adjusted
                     * to avoid overwriting of next warrior.
                     */
                    if (warrior.Count > minDistance + 1) 
                        right -= (warrior.Count - minDistance - 1);
                    /* If there is at least one address to load first instruction of warrior
                     * even after adjusting right boundary, gap is considered to be possible.
                     */
                    if (right - left >= 0)
                        possibleGaps.Add(i);
                    i++;
                }
                // Calculating load address.
                if (possibleGaps.Count == 0)
                    return ErrorState.MinDistance; // Core is full.
                else
                {
                    // Choosing random gap of possible gaps.
                    i = possibleGaps[randomizer.Next(0, possibleGaps.Count - 1)];
                    left = gaps[i].left;
                    right = gaps[i].right;
                    if (warrior.Count > minDistance + 1)
                        right -= (warrior.Count - minDistance - 1); // Adjusting right boundary.
                    /* First warrior is loaded to the first address of empty core. Any other
                     * next warrior is loaded to randomly chosen address of selected gap.
                     */
                    if (left == 0)
                        loadAddress = left; 
                    else
                        loadAddress = randomizer.Next(left, right);
                    right = gaps[i].right; // Right boundary must be de-adjusted before cutting gap.
                    gaps.RemoveAt(i);
                    /* If there is some remaining range of addresses on the left side of the cut
                     * and if it is relevant according to mindistance limit, new gap is added
                     * to gaps list.
                     */
                    if ((loadAddress - minDistance - 1) - left >= 0)
                        gaps.Add(new Gap(left, loadAddress - minDistance - 1));
                    /* If there is some remaining range of addresses on the right side of the cut
                     * and if it is relevant according to mindistance limit, new gap is added
                     * to gaps list. If warrior is longer than min distance, left boundary of
                     * new gap must be adjusted in order to avoid later overwriting of warrior.
                     */
                    if (warrior.Count > minDistance + 1)
                    {
                        if (right - (loadAddress + warrior.Count ) >= 0)
                            gaps.Add(new Gap(loadAddress + warrior.Count, right));
                    }
                    else
                    {
                        if (right - (loadAddress + minDistance + 1) >= 0)
                            gaps.Add(new Gap(loadAddress + minDistance + 1, right));
                    }            
                }
            }

            // Creating new warrior.
            Queues.Add(new Warrior(maxProcesses, Convert.ToByte(Queues.Count)));
            Queues[Queues.Count - 1].color = colors[(Queues.Count - 1) % colors.Length];
            
            // Loading warrior into core.
            for (int i = 0; i < warrior.Count; i++)
            {
                Core[(loadAddress + i) % coreSize] = warrior[i];
                // Each cell influenced by warrior is repainted by warrior's color.
                engineInterface.Repaint((loadAddress + i) % coreSize, Queues[Queues.Count - 1].color);
            }

            /* Warrior's task queue is initialized. Note that this must be done after loading
             * warrior to core in order to repaint first instruction in task queue by different 
             * color.
             */
            Queues[Queues.Count - 1].PushTask((loadAddress + startInstruction) % coreSize);

            return ErrorState.Normal; // success
        }

        /* Main executive function - Execute MARS Instrucion ICWS94.
         * First argument is current warrior, second argument is its program
         * counter - address of instruction to execute.
         */
        private ErrorState EMI94(Warrior warrior, Address PC)
        {
            /* Instruction register of currently executing instruction
             * referenced by program counter (PC).
             */
            Instruction IR;

            /* Instruction register for instruction referenced by 
             * A/B-operand.
             */
            Instruction IRA;
            Instruction IRB;

            /* Four following pointers are program counter (PC) relative.
             * Pointers are offsets to instruction referred to by the
             * A/B-operand for reading or writing.
             */
            Address RPA = 0;
            Address WPA = 0;
            Address RPB = 0;
            Address WPB = 0;

            /* Post-increment pointer is temporary address for 
             * post-increment indirect operands. Content of this address 
             * is incremented after indirect operand is successfully 
             * fetched from core.
             */
            Address PIP = 0;

            /* Current instuction is copied from core to Instruction
             * Register.
             */
            IR = Core[PC];

            // A-operand evaluation
            if (IR.aMode == Mode.Immediate)
                RPA = WPA = 0; // Immediate mode - operand points to itself.
            else
            {
                // Direct mode - operand's read/write offset is folded.
                RPA = Fold(IR.aNumber, readDistance);
                WPA = Fold(IR.aNumber, writeDistance);
                /* Indirect mode - A-operand is taken from intermediade instruction
                 * to calculate final destination address. Pre-decrementing or
                 * post-incrementing takes place depending on indirection mode.
                 */
                if (IR.aMode == Mode.A_indirect ||
                    IR.aMode == Mode.A_decrement ||
                    IR.aMode == Mode.A_increment)
                {
                    // Intermediate operand is pre-decremented.
                    if (IR.aMode == Mode.A_decrement)
                    {
                        Core[(PC + WPA) % coreSize].aNumber = (Core[(PC + WPA) % coreSize].aNumber + coreSize - 1) % coreSize;
                        //RPA = WPA;
                    }
                    // Intermediate operand address is memorized to post-increment it later.
                    else if (IR.aMode == Mode.A_increment)
                    {
                        PIP = (PC + WPA) % coreSize;
                        //RPA = WPA;
                    }
                    /* NOTE: Purpose of commented lines is to ensure the expected behaviour of
                     *       pre-decrementing - content of an address which is going to be read is
                     *       firstly decremented - and post-incrementing - content of an address
                     *       which is read is incremented after fetching. However this is in
                     *       conflict with ICWS94 standard and philosophy of read/write limit.
                     */
                    // Final offsets are folded again
                    RPA = Fold(RPA + Core[(PC + RPA) % coreSize].aNumber, readDistance);
                    WPA = Fold(WPA + Core[(PC + WPA) % coreSize].aNumber, writeDistance);
                }
                /* Indirect mode - B-operand is taken from intermediade instruction
                 * to calculate final destination address. Pre-decrementing or
                 * post-incrementing takes place depending on indirection mode.
                 */
                if (IR.aMode == Mode.B_indirect ||
                    IR.aMode == Mode.B_decrement ||
                    IR.aMode == Mode.B_increment)
                {
                    if (IR.aMode == Mode.B_decrement)
                    {
                        Core[(PC + WPA) % coreSize].bNumber = (Core[(PC + WPA) % coreSize].bNumber + coreSize - 1) % coreSize;
                        //RPA = WPA;
                    }
                    else if (IR.aMode == Mode.B_increment)
                    {
                        PIP = (PC + WPA) % coreSize;
                        //RPA = WPA;
                    }
                    RPA = Fold(RPA + Core[(PC + RPA) % coreSize].bNumber, readDistance);
                    WPA = Fold(WPA + Core[(PC + WPA) % coreSize].bNumber, writeDistance);
                }
            }
            /* Operand's target is already evaluated and can copied to 
             * corresponding instruction register.
             */
            IRA = Core[((PC + RPA) % coreSize)];
            /* Content of memorized address in Post-Increment Pointer (PIP)
             * can be finally incremented. Note that post-incrementing must
             * be last task in operand's evaluation.
             */
            if (IR.aMode == Mode.A_increment)
                Core[PIP].aNumber = (Core[PIP].aNumber + 1) % coreSize;
            else if (IR.aMode == Mode.B_increment)
                Core[PIP].bNumber = (Core[PIP].bNumber + 1) % coreSize;

            // B-operand evaluation
            if (IR.bMode == Mode.Immediate)
                RPB = WPB = 0;
            else
            {
                RPB = Fold(IR.bNumber, readDistance);
                WPB = Fold(IR.bNumber, writeDistance);
                if (IR.bMode == Mode.A_indirect ||
                    IR.bMode == Mode.A_decrement ||
                    IR.bMode == Mode.A_increment)
                {
                    if (IR.bMode == Mode.A_decrement)
                    {
                        Core[(PC + WPB) % coreSize].aNumber = (Core[(PC + WPB) % coreSize].aNumber + coreSize - 1) % coreSize;
                        //RPB = WPB;
                    }
                    else if (IR.bMode == Mode.A_increment)
                    {
                        PIP = (PC + WPB) % coreSize;
                        //RPB = WPB;
                    }
                    RPB = Fold(RPB + Core[(PC + RPB) % coreSize].aNumber, readDistance);
                    WPB = Fold(WPB + Core[(PC + WPB) % coreSize].aNumber, writeDistance);
                }
                if (IR.bMode == Mode.B_indirect ||
                    IR.bMode == Mode.B_decrement ||
                    IR.bMode == Mode.B_increment)
                {
                    if (IR.bMode == Mode.B_decrement)
                    {
                        Core[(PC + WPB) % coreSize].bNumber = (Core[(PC + WPB) % coreSize].bNumber + coreSize - 1) % coreSize;
                        //RPB = WPB;
                    }
                    else if (IR.aMode == Mode.B_increment)
                    {
                        PIP = (PC + WPB) % coreSize;
                        //RPB = WPB;
                    }
                    RPB = Fold(RPB + Core[(PC + RPB) % coreSize].bNumber, readDistance);
                    WPB = Fold(WPB + Core[(PC + WPB) % coreSize].bNumber, writeDistance);
                }
            }
            IRB = Core[((PC + RPB) % coreSize)];
            if (IR.bMode == Mode.A_increment)
                Core[PIP].aNumber = (Core[PIP].aNumber + 1) % coreSize;
            else if (IR.bMode == Mode.B_increment)
                Core[PIP].bNumber = (Core[PIP].bNumber + 1) % coreSize;

            // Both operands are evaluated. Instruction execution follows.
            switch (IR.opcode)
            {
                /* NOTE: For further explanation of used terminology, see 
                 *       ICWS94 standard or program documentation. Comments
                 *       of opcodes are slightly modified or same as explanations 
                 *       in ICWS94 standard to maintain clarity and comprehension.
                 *       Comments of modifiers are ommited, because their
                 *       behaviour is clear after reading the standard or 
                 *       documentation.
                 */

                /* Instruction with a DAT opcode have no further function.
                 * The current warrior's program counter is not updated and
                 * thus warrior looses one of its processes.
                 */
                case Opcode.DAT:
                    // GUI is reported of lost task.
                    engineInterface.TaskChange(warrior.id, false);
                    // If warrior is still alive, next instruction in task queue is repainted.
                    warrior.PaintExecute();
                    break;

                /* MOV replaces the B-target with the A-value and queues
                 * the next instruction.
                 */
                case Opcode.MOV:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                            Core[(PC + WPB) % coreSize].aNumber = IRA.aNumber;
                            break;
                        case Modifier.B:
                            Core[(PC + WPB) % coreSize].bNumber = IRA.bNumber;
                            break;
                        case Modifier.AB:
                            Core[(PC + WPB) % coreSize].bNumber = IRA.aNumber;
                            break;
                        case Modifier.BA:
                            Core[(PC + WPB) % coreSize].aNumber = IRA.bNumber;
                            break;
                        case Modifier.F:
                            Core[(PC + WPB) % coreSize].aNumber = IRA.aNumber;
                            Core[(PC + WPB) % coreSize].bNumber = IRA.bNumber;
                            break;
                        case Modifier.X:
                            Core[(PC + WPB) % coreSize].bNumber = IRA.aNumber;
                            Core[(PC + WPB) % coreSize].aNumber = IRA.bNumber;
                            break;
                        case Modifier.I:
                            Core[(PC + WPB) % coreSize] = IRA;
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    // Cell influenced by warrior is painted by its color.
                    engineInterface.Repaint((PC + WPB) % coreSize, warrior.color);
                    // Next instruction is added to warrior's task queue.
                    warrior.PushTask((PC + 1) % coreSize);
                    break;

                /* Arithmetic instructions replace the B-target with the result
                 * of corresponding operation above the A-value and B-value, 
                 * and queue the next instruction. Operation can be addition, 
                 * substraction or multiplication.
                 */
                case Opcode.ADD:
                case Opcode.SUB:
                case Opcode.MUL:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                            Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.aNumber, IR.opcode);
                            break;
                        case Modifier.B:
                            Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.bNumber, IR.opcode);
                            break;
                        case Modifier.AB:
                            Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.aNumber, IR.opcode);
                            break;
                        case Modifier.BA:
                            Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.bNumber, IR.opcode);
                            break;
                        case Modifier.I:
                        case Modifier.F:
                            Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.aNumber, IR.opcode);
                            Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.bNumber, IR.opcode);
                            break;
                        case Modifier.X:
                            Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.aNumber, IR.opcode);
                            Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.bNumber, IR.opcode);
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    // Cell influenced by warrior is painted by its color.
                    engineInterface.Repaint((PC + WPB) % coreSize, warrior.color);
                    // Next instruction is added to warrior's task queue.
                    warrior.PushTask((PC + 1) % coreSize);
                    break;
                
                /* DIV and MOD replace the B-target with the integral
                 * quotient (for DIV) or remainder (for MOD) of the B-value
                 * by the A-value, and queues the next instruction.
                 * Process is removed from warrior's task queue if A-value 
                 * is zero - same effect as execution of DAT.
                 */
                case Opcode.DIV:
                case Opcode.MOD:
                    bool divisionByZero = false;
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                            if (IRA.aNumber != 0)
                                Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.aNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            break;
                        case Modifier.B:
                            if (IRA.bNumber != 0)
                                Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.bNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            break;
                        case Modifier.AB:
                            if (IRA.aNumber != 0)
                                Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.aNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            break;
                        case Modifier.BA:
                            if (IRA.bNumber != 0)
                                Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.bNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            break;
                        case Modifier.I:
                        case Modifier.F:
                            if (IRA.aNumber != 0)
                                Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.aNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            if (IRA.bNumber != 0)
                                Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.bNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            break;
                        case Modifier.X:
                            if (IRA.aNumber != 0)
                                Core[(PC + WPB) % coreSize].bNumber = EMI94Calculator(IRB.bNumber, IRA.aNumber, IR.opcode);
                            else
                                divisionByZero = true;
                            if (IRA.bNumber != 0)
                                Core[(PC + WPB) % coreSize].aNumber = EMI94Calculator(IRB.aNumber, IRA.bNumber, IR.opcode);
                            else
                                divisionByZero = true; 
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    if (divisionByZero == false)
                    {
                        // Cell influenced by warrior is painted by its color.
                        engineInterface.Repaint((PC + WPB) % coreSize, warrior.color);
                        // Next instruction is added to warrior's task queue.
                        warrior.PushTask((PC + 1) % coreSize);
                    }
                    else
                    {
                        // GUI is reported of lost task.
                        engineInterface.TaskChange(warrior.id, false);
                        // If warrior is still alive, next instruction in task queue is repainted.
                        warrior.PaintExecute();
                    }
                    break;

                // JMP queues the sum of the program counter and the A-pointer.
                case Opcode.JMP:
                    warrior.PushTask((PC + RPA) % coreSize);
                    break;

                /* JMZ queues the sum of the program counter and A-pointer
                 * if the B-value is zero.  Otherwise, it queues the next instruction.
                 * JMN queues the sum of the Program Counter and A-Pointer
                 * if the B-value is not zero.  Otherwise, it queues the
                 * next instruction.
                 */
                case Opcode.JMZ:
                case Opcode.JMN:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                        case Modifier.BA:
                            if (EMI94Comparator(IRB.aNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.B:
                        case Modifier.AB:
                            if (EMI94Comparator(IRB.bNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.I:
                        case Modifier.F:
                        case Modifier.X:
                            if (EMI94Comparator(IRB.aNumber, 0, IR.opcode) && EMI94Comparator(IRB.bNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    break;

                /* DJN decrements the B-value and the B-target, then tests if 
                 * the B-value is zero.  If the result is not zero, the sum of 
                 * the program counter and A-pointer is queued.  Otherwise, the 
                 * next instruction is queued.
                 */
                case Opcode.DJN:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                        case Modifier.BA:
                            Core[(PC + WPB) % coreSize].aNumber = (Core[(PC + WPB) % coreSize].aNumber + coreSize -1) % coreSize;
                            IRB.aNumber = (IRB.aNumber + coreSize - 1) % coreSize; 
                            if (EMI94Comparator(IRB.aNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.B:
                        case Modifier.AB:
                            Core[(PC + WPB) % coreSize].bNumber = (Core[(PC + WPB) % coreSize].bNumber + coreSize - 1) % coreSize;
                            IRB.bNumber = (IRB.bNumber + coreSize - 1) % coreSize;
                            if (EMI94Comparator(IRB.bNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.I:
                        case Modifier.F:
                        case Modifier.X:
                            Core[(PC + WPB) % coreSize].aNumber = (Core[(PC + WPB) % coreSize].aNumber + coreSize - 1) % coreSize;
                            IRB.aNumber = (IRB.aNumber + coreSize - 1) % coreSize;
                            Core[(PC + WPB) % coreSize].bNumber = (Core[(PC + WPB) % coreSize].bNumber + coreSize - 1) % coreSize;
                            IRB.bNumber = (IRB.bNumber + coreSize - 1) % coreSize;
                            if (EMI94Comparator(IRB.aNumber, 0, IR.opcode) && EMI94Comparator(IRB.bNumber, 0, IR.opcode))
                                warrior.PushTask((PC + RPA) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    break;

                /* SEQ/CMP compares the A-value and the B-value. If there
                 * are no differences, then the instruction after the next
                 * instruction is queued.  Otherwise, the next instrution 
                 * is queued. 
                 * SNE compares the A-value and the B-value. If there
                 * is a difference, then the instruction after the next
                 * instruction is queued.  Otherwise, the next instrution 
                 * is queued.
                 */
                case Opcode.CMP:
                case Opcode.SNE:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                            if (EMI94Comparator(IRA.aNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.B:
                            if (EMI94Comparator(IRA.bNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.AB:
                            if (EMI94Comparator(IRA.aNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.BA:
                            if (EMI94Comparator(IRA.bNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.F:
                            if (EMI94Comparator(IRA.aNumber, IRB.aNumber, IR.opcode) &&
                                EMI94Comparator(IRA.bNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.X:
                            if (EMI94Comparator(IRA.aNumber, IRB.bNumber, IR.opcode) &&
                                EMI94Comparator(IRA.bNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.I:
                            if (EMI94InstructionComparator(IRA, IRB, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    break;

                /* SLT queues the instruction after the next instruction 
                 * if A-value is less than B-value. Otherwise, the next 
                 * instruction is queued.
                 */
                case Opcode.SLT:
                    switch (IR.modifier)
                    {
                        case Modifier.A:
                            if (EMI94Comparator(IRA.aNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.B:
                            if (EMI94Comparator(IRA.bNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.AB:
                            if (EMI94Comparator(IRA.aNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.BA:
                            if (EMI94Comparator(IRA.bNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.I:
                        case Modifier.F:
                            if (EMI94Comparator(IRA.aNumber, IRB.aNumber, IR.opcode) &&
                                EMI94Comparator(IRA.bNumber, IRB.bNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        case Modifier.X:
                            if (EMI94Comparator(IRA.aNumber, IRB.bNumber, IR.opcode) &&
                                EMI94Comparator(IRA.bNumber, IRB.aNumber, IR.opcode))
                                warrior.PushTask((PC + 2) % coreSize);
                            else
                                warrior.PushTask((PC + 1) % coreSize);
                            break;
                        default:
                            return ErrorState.UndefinedInstruction;
                    }
                    break;

                /* SPL queues the next instruction and also queues the sum 
                 * of the program counter and A-pointer. Note that if warrior's
                 * queue was full before execution of current task, there is
                 * space only for the first instruction. In that case the 
                 * behaviour of SPL is same as NOP.
                 */
                case Opcode.SPL:
                    warrior.PushTask((PC + 1) % coreSize);
                    if (warrior.TaskCount < maxProcesses)
                        // GUI is reported of new task.
                        engineInterface.TaskChange(warrior.id, true);
                    warrior.PushTask((PC + RPA) % coreSize);
                    break;

                // NOP queues the next instruction.
                case Opcode.NOP:
                    warrior.PushTask((PC + 1) % coreSize);
                    break;

                default:
                    return ErrorState.UndefinedInstruction;
            }
            // Cell influenced by warrior is painted by its color.
            engineInterface.Repaint(PC, warrior.color);
            return ErrorState.Normal; // success
        }
        /* Auxiliary function which provides arithmetic calculations to EMI94.
         * Note that result must be valid core address, thus some of the operations
         * must be modulo the size of core.
         */
        private Address EMI94Calculator(Address operand1, Address operand2, Opcode operation)
        {
            Address result = 0;
            switch (operation)
            {
                case Opcode.ADD:
                    result = (operand1 + operand2) % coreSize;
                    break;
                case Opcode.SUB:
                    result = (operand1 + coreSize - operand2) % coreSize;
                    break;
                case Opcode.MUL:
                    result = (operand1 * operand2) % coreSize;
                    break;
                case Opcode.DIV:
                    result = operand1 / operand2;
                    break;
                case Opcode.MOD:
                    result = operand1 % operand2;
                    break;
            }
            return result;
        }
        /* Auxiliary function which provides address comparation to EMI94.
         */
        private bool EMI94Comparator(Address operand1, Address operand2, Opcode operation)
        {
            bool result = false;
            switch (operation)
            {
                case Opcode.JMZ:
                case Opcode.CMP:
                    result = (operand1 == operand2);
                    break;
                case Opcode.JMN:
                case Opcode.DJN:
                case Opcode.SNE:
                    result = (operand1 != operand2);
                    break;
                case Opcode.SLT:
                    result = (operand1 < operand2);
                    break;
            }
            return result;
        }
        /* Auxiliary function which provides instruction comparation to EMI94.
         */
        private bool EMI94InstructionComparator(Instruction operand1, Instruction operand2, Opcode operation)
        {
            bool result = false;
            switch (operation)
            {
                case Opcode.CMP:
                    result = (operand1.opcode == operand2.opcode && operand1.modifier == operand2.modifier &&
                              operand1.aMode == operand2.aMode && operand1.bMode == operand2.bMode &&
                              operand1.aNumber == operand2.aNumber && operand1.bNumber == operand2.bNumber);
                    break;
                case Opcode.SNE:
                    result = (operand1.opcode != operand2.opcode && operand1.modifier != operand2.modifier &&
                              operand1.aMode != operand2.aMode && operand1.bMode != operand2.bMode &&
                              operand1.aNumber != operand2.aNumber && operand1.bNumber != operand2.bNumber);
                    break;
            }
            return result;
        }

        /* Function is designed to be running on different thread than GUI to ensure
         * that GUI stay responsive. Communication with GUI is done strictly through
         * engine interface. Function successively executes intructions of living
         * warriors in loop. Speed of the loop is determined by step mechanism and
         * speed adjusted in GUI.
         */
        private Int32 cycles; // number of executed cycles since last core wipe
        public void Start()
        {
            IsBusy = true;
            engineInterface.Started(); // Engine is busy from now on.
            byte warriorIndex = 0;
            ErrorState result = ErrorState.Normal;
            /* Battle cannot continue when there is only one alive warrior remaining 
             * or when maximum cycles limit is reached. Loop also halts when it is 
             * requested by step mechanism or by user. If error occurs within main
             * executive function, battle halts and GUI is requested to inform user.
             */
            while (Queues.Count > 1 && !backgroundWorker.CancellationPending && cycles < maxCycles)
            {
                Warrior warrior = Queues[warriorIndex];
                if (warrior.IsAlive())
                {
                    Address task = warrior.PopTask();
                    /* Thread is slowed down by sleep operation only if step mechanism
                     * is disabled and speed is set below maximum.
                     */
                    if (step == false && SleepLimit > 0)
                        Thread.Sleep(SleepLimit);
                    result = EMI94(warrior, task); // Execute current warrior's task.
                }
                else
                {
                    Queues.RemoveAt(warriorIndex); // Dead warrior is removed in order to speed up loop.
                    warriorIndex--;
                }
                warriorIndex = decimal.ToByte((warriorIndex + 1) % Queues.Count);
                /* If step mechanism is active and each warrior has executed one instruction,
                 * thread is halted and engine is stopped.
                 */
                if (step == true && warriorIndex == 0)
                {
                    backgroundWorker.CancelAsync();
                    IsBusy = false;
                    engineInterface.Finished();
                }
                if (warriorIndex == 0) // Each warrior has executed one instruction.
                {
                    cycles++;
                    engineInterface.Cycle(); // GUI is reported that cycle occured.
                }
                if (result != ErrorState.Normal)
                {
                    backgroundWorker.CancelAsync();
                    IsBusy = false;
                    engineInterface.Finished();
                    engineInterface.Error(result); // GUI is reported about error.
                }
            }
            IsBusy = false;
            engineInterface.Finished(); // Engine is idle from now on.
        }

        // Engine is initialized by its constructor.
        public Engine()
        {
            randomizer = new Random();
            backgroundWorker = new BackgroundWorker();
            backgroundWorker.WorkerSupportsCancellation = true;
            backgroundWorker.DoWork += new DoWorkEventHandler(this.backgroundWorker_DoWork);
            LoadDefault(); // default predefined variables set
            SleepLimit = 25; // default sleep limit (25ms)
            step = false;
            IsBusy = false;
        }
    }
}
