﻿/* CorWX - The Core War simulator
 * Copyright (c) 2009 Petr Koupy <petr.koupy@gmail.com>
 * 
 * File: Parser.cs
 * 
 * Description:
 * Redcode parser compatible with ICWS 94' Draft 3.3.
 * 
 * Revision history:
 * Rev  Date        Who             Description
 * 1.0  17/08/08    Petr Koupy      Initial draft
 * 
 * GNU General Public License Usage
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace CorWX
{
    /* Parser is responsible for decoding and parsing assembly file into load file
     * which is later handed over to engine. 
     */
    class Parser
    {
        // Engine instance provides opcodes, modifiers, modes and predefined variables.
        private Engine engine; 
        // Determines whether older assembly files should be considered as ICWS88 or ICWS86 files.
        static public bool conversion88 = true;

        // Parser is initialized by its constructor.
        public Parser(Engine engine)
        {
            this.engine = engine;
        }

        /* Assembly file gramatics
         * assembly_file:
         *        line+ EOF
         * line:
         *        comment | instruction
         * comment:
         *        ; v* newline | newline
         * instruction:
         *        label_list operation mode expr comment |
         *        label_list operation mode expr , mode expr comment
         * label_list:
         *        label newline* label_list | e
         * label:
         *        alpha alphanumeral*
         * operation:
         *        opcode | opcode.modifier
         * opcode:
         *        DAT | MOV | ADD | SUB | MUL | DIV | MOD |
         *        JMP | JMZ | JMN | DJN | CMP | SEQ | SNE |
         *        SLT | SPL | NOP | ORG | EQU | END
         * modifier:
         *        A | B | AB | BA | F | X | I
         * mode:
         *        # | $ | * | @ | { | < | } | > | e
         * expr:
         *        term |
         *        term + expr | term - expr |
         *        term * expr | term / expr |
         *        term % expr
         * term:
         *        label | number | ( expression )
         * number:
         *        whole_number | signed_integer
         * signed_integer:
         *        +whole_number | -whole_number
         * whole_number:
         *        numeral+
         * alpha:
         *        A-Z | a-z | _
         * numeral:
         *        0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
         * alphanumeral:
         *        alpha | numeral
         * v:
         *        ^newline
         * newline:
         *        LF | CR | LF CR | CR LF
         * e:
         */

        // Gramatics above is represented by tokens of listed kinds.
        private enum Kind 
        { 
            EOF,                // Peek() returns -1
            EOL,                // \n | \r | \n\r | \r\n
            Addition,           // +
            Substraction,       // -
            Multiplication,     // *
            Divison,            // /
            Modulo,             // %
            LeftParenthesis,    // (
            RightParenthesis,   // )
            Number,             // decimal number composed by numerals 0-9
            Label,              // string of A-Z | a-z | _ | 0-9 (first character must not be number)
            PseudoOpcode,       // ORG | EQU | END
            Opcode,             // DAT | MOV | ADD | SUB | MUL | DIV | MOD | JMP | JMZ | JMN | DJN | CMP | SEQ | SNE | SLT | SPL | NOP
            Modifier,           // A | B | AB | BA | F | X | I
            Mode,               // # | $ | * | @ | { | < | } | >
            Comma,              // ,
        }

        /* Static array of modes is used to determine index into engine array of modes
         * in order to provide translation between listed characters and corresponding
         * mode representations within engine. Order of elements array must correlate 
         * with similar structures in engine.
         */
        public static readonly char[] modes = new char[] { '#', '$', '*', '@', '{', '<', '}', '>' };

        /* For further explanation of instruction pseudo opcodes see ICWS94 3.3 on line 0270.
         * Static array is used for searching specific pseudo opcode during parsing.
         */
        public static readonly PseudoOpcode[] pseudoOpcode = new PseudoOpcode[] 
        { 
          PseudoOpcode.ORG, PseudoOpcode.EQU, PseudoOpcode.END
        }; 
        public enum PseudoOpcode
        {
            ORG, // ORiGin
            EQU, // EQUate
            END  // ENDing
        }

        /* ICWS88 to ICWS94 conversion table
         Opcode                     A-mode     B-mode     modifier
         ---------------------------------------------------------
         DAT                        #$@<>*{}   #$@<>*{}   F
         MOV,CMP,SEQ,SNE            #          #$@<>*{}   AB
                                    $@<>*{}    #          B
                                    $@<>*{}    $@<>*{}    I
         ADD,SUB,MUL,DIV,MOD        #          #$@<>*{}   AB
                                    $@<>*{}    #          B
                                    $@<>*{}    $@<>*{}    F
         SLT                        #          #$@<>*{}   AB
                                    $@<>*{}    #$@<>*{}   B
         JMP,JMZ,JMN,DJN,SPL,NOP    #$@<>*{}   #$@<>*{}   B
         ---------------------------------------------------------
         */
        private string Convert88to94(string opcode, string aMode, string bMode)
        {
            switch (opcode)
            {
                case "DAT":
                    return ".F";
                case "MOV":
                case "CMP":
                case "SEQ":
                case "SNE":
                    if (aMode == "#")
                        return ".AB";
                    else if (bMode == "#")
                        return ".B";
                    else
                        return ".I";
                case "ADD":
                case "SUB":
                case "MUL":
                case "DIV":
                case "MOD":
                    if (aMode == "#")
                        return ".AB";
                    else if (bMode == "#")
                        return ".B";
                    else
                        return ".F";
                case "SLT":
                    if (aMode == "#")
                        return ".AB";
                    else
                        return ".B";
                default:
                    return ".B";
            }
        }

        /* ICWS86 to ICWS94 conversion table
         Opcode                  A-mode     B-mode     modifier
         ---------------------------------------------------------
         DAT                        #$@<>*{}   #$@<>*{}   F
         MOV,CMP,SEQ,SNE            #          #$@<>*{}   AB
                                    $@<>*{}    #          B
                                    $@<>*{}    $@<>*{}    I
         ADD,SUB,MUL,DIV,MOD        #          #$@<>*{}   AB
                                    $@<>*{}    #$@<>*{}   B
         SLT                        #          #$@<>*{}   AB
                                    $@<>*{}    #$@<>*{}   B
         JMP,JMZ,JMN,DJN,SPL,NOP    #$@<>*{}   #$@<>*{}   B
         ---------------------------------------------------------
         */
        private string Convert86to94(string opcode, string aMode, string bMode)
        {
            switch (opcode)
            {
                case "DAT":
                    return ".F";
                case "MOV":
                case "CMP":
                case "SEQ":
                case "SNE":
                    if (aMode == "#")
                        return ".AB";
                    else if (bMode == "#")
                        return ".B";
                    else
                        return ".I";
                case "ADD":
                case "SUB":
                case "MUL":
                case "DIV":
                case "MOD":
                case "SLT":
                    if (aMode == "#")
                        return ".AB";
                    else
                        return ".B";
                default:
                    return ".B";
            }
        }

        /* Token is elementary unit of assembly file which is created by 
         * scanner and recognized by parser. There are several kinds of
         * tokens (see above). Some tokens are determined directly by their
         * kind. Other tokens are containers for decimal value or string or
         * both.
         */
        private struct Token 
        {
            public readonly Kind kind;
            public readonly Int32 value;
            public readonly string label;
            
            private Token(Int32 value) 
            {
                this.kind = Kind.Number;
                this.value = value;
                this.label = "";
            }

            private Token(Kind kind, string label)
            {
                this.kind = kind;
                this.value = 0;
                this.label = label;
            }

            public override string ToString() 
            {
                if (kind == Kind.Number)
                    return value.ToString();
                else
                    return label;
            }

            static public Token FromValue(Int32 value) 
            {
                return new Token(value);
            }

            static public Token FromLabel(Kind kind, string label)
            {
                return new Token(kind, label);
            }
        }

        /* Token stream is responsible for conversion of file into stream of
         * tokens. Each time parser needs another token, token stream provides
         * it by embedded scanner.
         */
        private class TokenStream
        {
            // Token stream is tied with some assembly file.
            private StreamReader reader;

            // Token stream remembers only last scanned token.
            private Token currentToken;
            public Token CurrentToken
            {
                get
                {
                    return currentToken;
                }
            }

            public TokenStream(StreamReader reader)
            {
                this.reader = reader;
                currentToken = Token.FromLabel(Kind.EOF, "EOF"); // initialization
                this.Next(); // first token of file
            }

            /* Updates current token with new token provided by scanner on demand.
             * New token is also returned as return value.
             */
            public Token Next()
            {
                currentToken = Scan(reader);
                return currentToken;
            }
            
            // Determines whether given character represents mode.
            private bool IsMode(char test)
            {
                bool result = false;
                foreach (char mode in modes)
                    if (mode == test)
                        result = true;
                return result;
            }

            /* Reads comment from input stream and throws it away. Comment is terminated
             * by newline.
             */
            private void ScanComment(StreamReader reader)
            {
                while ((char)reader.Peek() != '\r' && (char)reader.Peek() != '\n')
                    reader.Read();
            }

            // Reads alphanumeric string from input stream and returns it.
            private string ScanWord(StreamReader reader)
            {
                StringBuilder builder = new StringBuilder();
                while (Char.IsLetterOrDigit((char)reader.Peek()) || (char)reader.Peek() == '_')
                    builder.Append((char)reader.Read());
                return builder.ToString();
            }

            // Reads numeric string from input stream and returns it as a decimal value.
            private Int32 ScanNumber(TextReader reader) 
            {
                Int32 result = reader.Read() - '0';
                while (Char.IsDigit((char)reader.Peek()))
                    result = 10 * result + reader.Read() - '0';
                return result;
            }

            /* Main scan function reads input stream until new token is recognized,
             * created and returned.
             */
            private Token Scan(StreamReader reader) 
            {
                while (reader.Peek() != -1) 
                {
                    // End of line is detected and returned. Other white spaces are thrown away.
                    if (Char.IsWhiteSpace((char)reader.Peek()))
                    {
                        char white = (char)reader.Read();
                        if (white == '\r')
                        {
                            if ((char)reader.Peek() == '\n')
                                reader.Read();
                            if (currentToken.kind != Kind.EOL)
                                return Token.FromLabel(Kind.EOL, "\r\n");
                        }
                        else if (white == '\n')
                        {
                            if ((char)reader.Peek() == '\r')
                                reader.Read();
                            if (currentToken.kind != Kind.EOL)
                                return Token.FromLabel(Kind.EOL, "\r\n");
                        }
                    }
                    // Semicolon determines start of comment. Comment is thrown away by auxiliary function.
                    else if ((char)reader.Peek() == ';')
                    {
                        ScanComment(reader);
                    }
                    // Alphabetic character ot '_' determines start of label or (pseudo)opcode.
                    else if (Char.IsLetter((char)reader.Peek()) || (char)reader.Peek() == '_')
                    {
                        string word = ScanWord(reader);
                        if (word.Length == 3) // It is possible that word is (pseudo)opcode.
                        {
                            for (int i = 0; i < Engine.opcode.Length; i++) // Opcodes are examined.
                            {
                                if (Engine.opcode[i].ToString("G") == word.ToUpper())
                                    return Token.FromLabel(Kind.Opcode, word.ToUpper());
                            }
                            if (word.ToUpper() == "SEQ") // SEQ is not opcode. It is only alias for CMP opcode.
                                return Token.FromLabel(Kind.Opcode, "CMP");
                            for (int i = 0; i < pseudoOpcode.Length; i++) // Pseudo opcodes are examined.
                            {
                                if (pseudoOpcode[i].ToString("G") == word.ToUpper())
                                    return Token.FromLabel(Kind.PseudoOpcode, word.ToUpper());
                            }
                        }
                        return Token.FromLabel(Kind.Label, word); // Word is label.
                    }
                    // Dot determines start of modifier.
                    else if ((char)reader.Peek() == '.')
                    {
                        reader.Read(); // Dot is thrown away
                        string word = ScanWord(reader);
                        for (int i = 0; i < Engine.modifier.Length; i++) // Modifiers are examined.
                        {
                            if (Engine.modifier[i].ToString("G") == word.ToUpper())
                                return Token.FromLabel(Kind.Modifier, "." + word.ToUpper());
                        }
                    }
                    /* Character is examined as a mode character. Last token must be controlled in 
                     * order to distinguish '*' A-indirect mode mode from '*' multiplication.
                     */
                    else if (IsMode((char)reader.Peek()) &&
                             (currentToken.kind == Kind.Opcode || currentToken.kind == Kind.Modifier || currentToken.kind == Kind.Comma))
                    {
                        return Token.FromLabel(Kind.Mode, char.ToString((char)reader.Read()));
                    }
                    // Comma is detected directly.
                    else if ((char)reader.Peek() == ',')
                    {
                        return Token.FromLabel(Kind.Comma, char.ToString((char)reader.Read()));
                    }
                    // Numeral determines start of decimal number.
                    else if (Char.IsDigit((char)reader.Peek()))
                    {
                        return Token.FromValue(ScanNumber(reader));
                    }
                    // Other cases are detected directly.
                    else
                    {
                        char operation = (char)reader.Read();
                        switch (operation)
                        {
                            case '+':
                                return Token.FromLabel(Kind.Addition, "+");
                            case '-':
                                return Token.FromLabel(Kind.Substraction, "-");
                            case '*':
                                return Token.FromLabel(Kind.Multiplication, "*");
                            case '/':
                                return Token.FromLabel(Kind.Divison, "/");
                            case '%':
                                return Token.FromLabel(Kind.Modulo, "%");
                            case '(':
                                return Token.FromLabel(Kind.LeftParenthesis, "(");
                            case ')':
                                return Token.FromLabel(Kind.RightParenthesis, ")");
                            default:
                                throw new ApplicationException("Parse error: illegal character");
                        }
                    }
                }
               return Token.FromLabel(Kind.EOF, "EOF");
            }
        }

        /* Main parser function. Parsing is done in five passes.
         * Input is assembly file, output is load file.
         * Note that all exceptions generated by parser must be
         * catched by caller of this public function.
         */
        public Stream Parse(StreamReader streamReader)
        {
            Hashtable labelTable = new Hashtable();
            List<string> labelStack = new List<string>();

            StreamReader reader = streamReader;
            TokenStream tokenStream = new TokenStream(reader);
            Stream stream = FirstPass(tokenStream, labelTable, labelStack);
            reader = new StreamReader(stream);
            tokenStream = new TokenStream(reader);
            stream = SecondPass(tokenStream, labelTable);
            reader = new StreamReader(stream);
            tokenStream = new TokenStream(reader);
            stream = ThirdPass(tokenStream, labelTable, labelStack);
            reader = new StreamReader(stream);
            tokenStream = new TokenStream(reader);
            stream = FourthPass(tokenStream, labelTable);
            reader = new StreamReader(stream);
            tokenStream = new TokenStream(reader);
            stream = FifthPass(tokenStream);
            return stream;
        }

        /* First pass
         * EQU labels and substitution strings are collected into hash table.
         */
        private Stream FirstPass(TokenStream tokenStream, Hashtable equTable, List<string> labelStack)
        {
            List<string[]> equList = new List<string[]>(); // temporary store for labels and their substitution strings
            Stream result = new MemoryStream();
            StreamWriter writer = new StreamWriter(result); // output stream

            /* Reserved labels for engine predefined variables are added into list.
             * Note that this must be done before any other label is added into list
             * because other labels might be using predefined variables in their 
             * substitution strings.
             */
            equList.Add(new string[2] { "CORESIZE", engine.CoreSize.ToString() });
            equList.Add(new string[2] { "MAXCYCLES", engine.MaxCycles.ToString() });
            equList.Add(new string[2] { "MAXLENGTH", engine.MaxLength.ToString() });
            equList.Add(new string[2] { "MAXPROCESSES", engine.MaxProcesses.ToString() });
            equList.Add(new string[2] { "MINDISTANCE", engine.MinDistance.ToString() });

            // Token stream is analyzed.
            while (tokenStream.CurrentToken.kind != Kind.EOF)
            {
                switch (tokenStream.CurrentToken.kind)
                {
                    case Kind.Label:
                        // Multiple labels are added into stack until (pseudo) opcode appears.
                        labelStack.Add(tokenStream.CurrentToken.ToString());
                        break;
                    case Kind.PseudoOpcode:
                        if (tokenStream.CurrentToken.label == "EQU")
                        {
                            // Pseudo opcode is EQU, line is read and added into list for each label in stack.
                            StringBuilder builder = new StringBuilder();
                            tokenStream.Next(); // 'EQU' is thrown away
                            while (tokenStream.CurrentToken.kind != Kind.EOL && tokenStream.CurrentToken.kind != Kind.EOF)
                            {
                                builder.Append(" " + tokenStream.CurrentToken.ToString() + " ");
                                tokenStream.Next();
                            }
                            foreach (string label in labelStack)
                            {
                                string[] equ = new string[2];
                                equ[0] = label;
                                equ[1] = builder.ToString();
                                equList.Add(equ);
                            }
                            // All labels are processed, stack can be wiped.
                            labelStack.Clear(); 
                        }
                        else
                        {
                            // Pseudo opcode is not EQU, line is read and write directly to output along with pseudo opcode.
                            while (tokenStream.CurrentToken.kind != Kind.EOL && tokenStream.CurrentToken.kind != Kind.EOF)
                            {
                                writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                                tokenStream.Next();
                            }
                            writer.Write("\r\n");
                            // Label stack should be empty. If not, labels are considered invalid and thrown away.
                            labelStack.Clear();
                        }
                        break;
                    case Kind.Opcode:
                        /* Token is opcode. Collected labels are written directly to output along with opcode and
                         * rest of the line.
                         */
                        foreach (string label in labelStack)
                            writer.Write(" " + label + " ");
                        labelStack.Clear();
                        while (tokenStream.CurrentToken.kind != Kind.EOL && tokenStream.CurrentToken.kind != Kind.EOF)
                        {
                            writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                            tokenStream.Next();
                        }
                        writer.Write("\r\n");
                        break;
                    default:
                        // Other lines are written directly to output. 
                        writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                        break;
                }
                tokenStream.Next();
            }
            // Label stack should be empty after EOF detected.
            if (labelStack.Count != 0)
                throw new ApplicationException("Parse error: opcode expected");

            /* EQUs can be nested in same order as their appearance in file. In order
             * to evaulate final substitution strings, it is needed to subsitute labels
             * in strings within each other.
             */
            for (int i = 0; i < equList.Count; i++)
                for (int j = 0; j < equList.Count; j++)
                {
                    StringBuilder builder = new StringBuilder(equList[j][1]);
                    builder.Replace(equList[i][0], equList[i][1]);
                    equList[j][1] = builder.ToString();
                }
            
            // Labels and their final substitution strings can be added into hash table.
            for (int i = 0; i < equList.Count; i++)
                equTable.Add(equList[i][0], equList[i][1]);

            writer.Flush();
            result.Seek(0, SeekOrigin.Begin); // Output is rewinded.
            return result;
        }

        /* Second pass
         * EQU labels are substitued by their substitution strings.
         */
        private Stream SecondPass(TokenStream tokenStream, Hashtable equTable)
        {
            Stream result = new MemoryStream();
            StreamWriter writer = new StreamWriter(result); // output stream
            // Token stream is analyzed.
            while (tokenStream.CurrentToken.kind != Kind.EOF)
            {
                if (tokenStream.CurrentToken.kind == Kind.Label && equTable.Contains(tokenStream.CurrentToken.ToString()))
                    writer.Write(" " + equTable[tokenStream.CurrentToken.ToString()] + " ");
                else
                    writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                tokenStream.Next();
            }
            writer.Write("\r\n"); // Safe precaution for the third pass - instruction line must not end with EOF.
            equTable.Clear(); // Hash table is wiped for later use.
            writer.Flush();
            result.Seek(0, SeekOrigin.Begin); // Output is rewinded.
            return result;
        }

        /* Third pass
         * Instruction labels are collected. Missing modifiers, modes and operands are added
         * according to conversion table and definitions in standard.
         */
        private Stream ThirdPass(TokenStream tokenStream, Hashtable labelTable, List<string> labelStack)
        {
            string opcode = "";
            string modifier = "";
            string aMode = "";
            StringBuilder aNumber = new StringBuilder();
            string bMode = "";
            StringBuilder bNumber = new StringBuilder();
            string comma = "";
            int instructionLine = 0; // Incremented after each line that contains instruction is processed.
            Stream result = new MemoryStream();
            StreamWriter writer = new StreamWriter(result); // output stream
            // Token stream is analyzed.
            while (tokenStream.CurrentToken.kind != Kind.EOF)
            {
                switch (tokenStream.CurrentToken.kind)
                {
                    case Kind.Label:
                        // Labels before opcodes are added into stack. Label after opcodes are parts of operands.
                        if (opcode == "")
                            labelStack.Add(tokenStream.CurrentToken.ToString());
                        else
                        {
                            if (comma == "") // Comma has not been read yet.
                                aNumber.Append(" " + tokenStream.CurrentToken.ToString() + " ");
                            else
                                bNumber.Append(" " + tokenStream.CurrentToken.ToString() + " ");
                        }
                        break;
                    case Kind.PseudoOpcode:
                        // Line is read and write directly to output along with pseudo opcode.
                        while (tokenStream.CurrentToken.kind != Kind.EOL && tokenStream.CurrentToken.kind != Kind.EOF)
                        {
                            writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                            tokenStream.Next();
                        }
                        writer.Write("\r\n");
                        break;
                    case Kind.Opcode:
                        // Collected labels are added into hash table with current instruction line.
                        foreach (string label in labelStack)
                            labelTable.Add(label, instructionLine);
                        instructionLine++;
                        // All labels are processed, stack can be wiped.
                        labelStack.Clear();
                        // Opcode is recognized and saved.
                        opcode = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Modifier:
                        // Modifier is recognized and saved.
                        modifier = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Mode:
                        // Mode is recognized and saved.
                        if (comma == "") // Comma has not been read yet.
                            aMode = tokenStream.CurrentToken.ToString();
                        else
                            bMode = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Comma:
                        // Comma is recognized and saved. 
                        comma = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.EOL:
                        /* Whenever end of line is read, if it was instruction line,
                         * all informations to expand or convert instruction should
                         * be known. Completed instruction is written to output.
                         */
                        if (opcode != "") // Line should contain whole instruction.
                        {
                            if (aNumber.ToString() == "")
                                throw new ApplicationException("Parse error: operand expected");
                            if (aMode == "")
                                aMode = "$"; // Direct mode is default if ommited.
                            if (comma == "")
                                comma = ","; // Missing comma is added.
                            /* Some instructions need only one operand, thus second operand can be omitted.
                             * Omitted operand is considered to be B-operand. Only exception is DAT
                             * instruction where omitted operand is considered to be A-operand.
                             */
                            if (bNumber.ToString() == "")
                            {
                                // If whole operand is omitted, default mode is immediate and value is set to zero.
                                bMode = "#";
                                bNumber.Append("0");
                                if (opcode == "DAT")
                                {
                                    bMode = aMode;
                                    bNumber.Replace(bNumber.ToString(), aNumber.ToString());
                                    aMode = "#";
                                    aNumber.Replace(aNumber.ToString(), "0");
                                }
                            }
                            if (bMode == "")
                                bMode = "$"; // Direct mode is default if ommited.
                            if (modifier == "")
                            {
                                // Unknown modifier is determined according to conversion settings.
                                if (conversion88 == true)
                                    modifier = Convert88to94(opcode, aMode, bMode);
                                else
                                    modifier = Convert86to94(opcode, aMode, bMode);
                            }
                            // Expanded and converted instruction is written to output.
                            writer.Write(opcode + modifier + " " + aMode + aNumber.ToString() + comma + " " + bMode + bNumber.ToString() + "\r\n");
                            // Variables are reset for next instruction.
                            opcode = ""; modifier = ""; aMode = ""; aNumber.Remove(0, aNumber.Length); comma = "";  bMode = ""; bNumber.Remove(0, bNumber.Length);
                        }
                        else // Line does not contain instruction.
                            writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                        break;
                    default:
                        if (opcode != "" && comma == "") // Tokens are part of A-operand.
                            aNumber.Append(" " + tokenStream.CurrentToken.ToString() + " ");
                        else if (opcode != "" && comma != "") // Tokens are part of B-operand.
                            bNumber.Append(" " + tokenStream.CurrentToken.ToString() + " ");
                        else // Other tokens are written directly to output.
                            writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                        break;
                }
                tokenStream.Next();
            }
            writer.Flush();
            result.Seek(0, SeekOrigin.Begin); // Output is rewinded.
            return result;
        }

        /* Fourth pass
         * Operands are evaluated. Labels are substitued during evaluation.
         * Remaining pseudo opcodes (ORD, END) are detected, evaluated and removed.
         */
        private Stream FourthPass(TokenStream tokenStream, Hashtable labelTable)
        {
            int startInstruction = 0; // Number of line containing initial instruction of warrior.
            /* Incremented when line that contains instruction is detected. Value is provided
             * to expression calculator to determine final substitution value for label.
             */
            int instructionLine = -1; 
            // If true, evaluated expression is ORG/END value instead of normal instruction operand.
            bool pseudo = false;
            Stream result = new MemoryStream();
            StreamWriter writer = new StreamWriter(result); // output stream
            // Token stream is analyzed.
            while (tokenStream.CurrentToken.kind != Kind.EOF)
            { 
                switch (tokenStream.CurrentToken.kind)
                {
                    case Kind.PseudoOpcode:
                        /* Opcode itself is thrown away. Next evaluated expression will be considered
                         * as number of line containing initial instruction of warrior.
                         */
                        pseudo = true;
                        tokenStream.Next();
                        break;
                    case Kind.Opcode:
                        /* Opcode itself is written directly to output. Next evaluated expression 
                         * will be considered as operand value.
                         */
                        pseudo = false;
                        instructionLine++;
                        writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                        tokenStream.Next();
                        break;
                    case Kind.Addition:
                    case Kind.Substraction:
                    case Kind.Label:
                    case Kind.LeftParenthesis:
                    case Kind.Number:
                        // Expression start detected.
                        if (pseudo == true)
                            // Number of line containing initial instruction of warrior is stored in variable.
                            startInstruction = Expression(tokenStream, labelTable, 0);
                        else
                        {
                            // Operand value is written directly to output.
                            Int32 value = Expression(tokenStream, labelTable, instructionLine);
                            writer.Write(" " + value.ToString() + " ");
                        }
                        switch (tokenStream.CurrentToken.kind)
                        {
                            // Expression function reads next token. It should be comma, EOL or EOF.
                            case Kind.Comma:
                            case Kind.EOL:
                            case Kind.EOF:
                                break;
                            default:
                                throw new ApplicationException("Parse error: comma/EOL/EOF expected");
                        }
                        break;
                    default:
                        // Other tokens are written directly to output.
                        writer.Write(" " + tokenStream.CurrentToken.ToString() + " ");
                        tokenStream.Next();
                        break;
                }
            }
            labelTable.Clear(); // Hash table is wiped.
            // '%' token is exploited to mark last line of file containing initial instruction number.
            writer.Write("\r\n" + "%" + startInstruction.ToString()); 
            writer.Flush();
            result.Seek(0, SeekOrigin.Begin); // Output is rewinded.
            return result;
        }

        /* Fifth pass
         * Space and newline cleaning.
         */
        private Stream FifthPass(TokenStream tokenStream)
        {
            int sign = 1; // Indicates whether operand is positive or negative.
            string opcode = "";
            string modifier = "";
            string aMode = "";
            string aNumber = "";
            string comma = "";
            string bMode = "";
            string bNumber = "";
            Stream result = new MemoryStream();
            StreamWriter writer = new StreamWriter(result); // output stream
            // Token stream is analyzed.
            while (tokenStream.CurrentToken.kind != Kind.EOF)
            {
                switch (tokenStream.CurrentToken.kind)
                {
                    case Kind.Opcode:
                        opcode = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Modifier:
                        modifier = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Mode:
                        if (comma == "") // Comma has not been read yet.
                            aMode = tokenStream.CurrentToken.ToString();
                        else
                            bMode = tokenStream.CurrentToken.ToString();
                        break;
                    case Kind.Substraction:
                        sign = -1; // Following operand is negative.
                        break;
                    case Kind.Number:
                        if (aNumber == "")
                            aNumber = (sign * tokenStream.CurrentToken.value).ToString();
                        else
                            bNumber = (sign * tokenStream.CurrentToken.value).ToString();
                        sign = 1; // '+' character is obviously omitted before positive number.
                        break;
                    case Kind.EOL:
                        // Each line containing instruction is written to output. Other lines are thrown away.
                        if (opcode != "")
                        {
                            // Instruction is written directly to output.
                            writer.Write(opcode + modifier + " " + aMode + aNumber + comma + " " + bMode + bNumber + "\r\n");
                            // Variables are reset for next instruction.
                            opcode = ""; modifier = ""; aMode = ""; aNumber = ""; comma = "";  bMode = ""; bNumber = "";
                        }
                        break;
                    case Kind.Modulo:
                        /* Exploited identification of initial instruction number is thrown away.
                         * Only value itself is written to output as the last line of load file.
                         */
                        writer.Write(tokenStream.Next().value.ToString());
                        break;
                    case Kind.Comma:
                        comma = ",";
                        break;
                    default:
                        // Other tokens should be removed by previous passes.
                        throw new ApplicationException("Parse error: wrong syntax");
                }
                tokenStream.Next();
            }
            writer.Flush();
            result.Seek(0, SeekOrigin.Begin); // Output is rewinded.
            return result;
        }

        /* Expression calculator. Each expression is composed of terms. Each term is composed
         * of factors. Factor can containt parenthesis with nested expression. Factor is number
         * or label. Labels are substitued by given hash table. Input is read from token stream.
         * Current instruction line is handed over in order to evaluate final values of labels
         * during substitution.
         */
        private Int32 Expression(TokenStream tokenStream, Hashtable labelTable, int instructionLine) 
        {
            return ExpressionOperation(Term(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine); ;
        }
        private Int32 ExpressionOperation(Int32 value, TokenStream tokenStream, Hashtable labelTable, int instructionLine)
        {
            switch (tokenStream.CurrentToken.kind)
            {
                case Kind.Addition:
                    tokenStream.Next();
                    return ExpressionOperation(value + Term(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine);
                case Kind.Substraction:
                    tokenStream.Next();
                    return ExpressionOperation(value - Term(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine);
                default:
                    return value;
            }
        }
        private Int32 Term(TokenStream tokenStream, Hashtable labelTable, int instructionLine) 
        {
            return TermOperation(Factor(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine); 
        }
        private Int32 TermOperation(Int32 value, TokenStream tokenStream, Hashtable labelTable, int instructionLine)
        {
            switch (tokenStream.CurrentToken.kind)
            {
                case Kind.Multiplication:
                    tokenStream.Next();
                    return TermOperation(value * Factor(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine);
                case Kind.Divison: // Division by zero exception should be catched along with other exceptions.
                    tokenStream.Next();
                    return TermOperation(value / Factor(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine);
                case Kind.Modulo: // Division by zero exception should be catched along with other exceptions.
                    tokenStream.Next();
                    return TermOperation(value % Factor(tokenStream, labelTable, instructionLine), tokenStream, labelTable, instructionLine);
                default:
                    return value;
            }
        }
        private Int32 Factor(TokenStream tokenStream, Hashtable labelTable, int instructionLine)
        {
            int sign = 1; // Indicates whether factor is positive or negative.
            Int32 value = 0;
            if (tokenStream.CurrentToken.kind == Kind.Addition)
            {
                tokenStream.Next();
                sign = 1;
            }
            else if (tokenStream.CurrentToken.kind == Kind.Substraction)
            {
                tokenStream.Next();
                sign = -1;
            }
            switch (tokenStream.CurrentToken.kind)
            {
                case Kind.Label:
                    if (labelTable.Contains(tokenStream.CurrentToken.ToString()))
                        value = Int32.Parse(labelTable[tokenStream.CurrentToken.ToString()].ToString());
                    else
                        throw new ApplicationException("Parse error: label not assigned");
                    value -= instructionLine; // Offset from current line to line which is represented by label.
                    tokenStream.Next();
                    return sign * value;
                case Kind.Number:
                    value = tokenStream.CurrentToken.value;
                    tokenStream.Next(); 
                    return sign * value;
                case Kind.LeftParenthesis:
                    tokenStream.Next();
                    value = Expression(tokenStream, labelTable, instructionLine);
                    // Nested expression must be closed by right parenthesis.
                    if (tokenStream.CurrentToken.kind != Kind.RightParenthesis)
                        throw new ApplicationException("Parse error: expected ')'");
                    tokenStream.Next();
                    return sign * value;
                default:
                    throw new ApplicationException("Parse error: expected number or '('");
            }
        }
    }
}
