﻿/* CorWX - The Core War simulator
 * Copyright (c) 2009 Petr Koupy <petr.koupy@gmail.com>
 * 
 * File: WindowMain.xaml.cs
 * 
 * Description:
 * Graphical core, warrior management, battle management.
 * 
 * Revision history:
 * Rev  Date        Who             Description
 * 1.0  01/05/09    Petr Koupy      Initial draft
 *
 * GNU General Public License Usage
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Threading;
using System.Text;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

/* Address alias for easy modification of core size range.
 * Should be same as in engine.
 */
using Address = System.Int32;

namespace CorWX
{
    /* Main window contains graphical view of core and user interface.
     * User can add warriors, view their source code, view core contents,
     * setup core and manage speed of battle.
     */
    public partial class WindowMain : Window, EngineInterface
    {
        /* Size of squared core container must be divisible by all intended
         * sizes of one squared core cell. Current setup gives following
         * possibilites:
         * CellSize      SideCount       CellCount
         *        5             96            9216
         *        4            120           14400
         *        3            160           25600
         *        2            240           57600
         *        1            480          230400
         */
        static private int coreContainerSize = 480;
        static private int coreCellSize = 5;
        static private int coreContainerSideCount = coreContainerSize / coreCellSize;
        static private int coreCellCount = coreContainerSideCount * coreContainerSideCount;

        static private double DPIx;
        static private double DPIy;
        static private double penWidth = 0;
        static private bool rendering = true;

        static private Color coreContainerColor = Colors.MidnightBlue; // Background color of whole core container.
        static private Color coreCellColor = Colors.Black; // Color of empty core cell.
        static private Color coreViewEmptyColor = Colors.LightGray; // Core view box font color of instruction in empty cell.
        static private Color coreViewColor = Colors.Red; // Color of optional core view range slider.

        private Address coreViewAddress; // Address of first instruction to show in core view box.
        static private int coreViewCount = 10; // Number of instructions stored in core view box.

        private RenderTargetBitmap coreMap; // At each time stores actual color state of all cells in core.
        private Engine engine; // Instance of engine.
        private WindowWarrior windowWarrior; // Optional windows for viewing warrior's assembly and load file.

        private List<Color> coreColors; // Cell colors indexed by address.
        private List<Button> buttonList; // Each warrior has its own colored button to show form with source code.
        private List<ProgressBar> progressBarList; // Each warrior has its own progress bar to show load of its task queue.
        private List<Label> labelList; // Each warrior has its own label to show exact number of tasks in its task queue.
        private List<Label> coreViewList; // Stores instruction labels of core view box.
        private List<string> sourceList; // Assembly file of each warrior is stored to be shown on demand.
        private List<string> loadList; // Load file of each warrior is stored to be shown on demand.

        public WindowMain()
        {
            this.engine = new Engine();
            InitializeComponent();
            
            /* Core view box consists of labels that contains instruction strings loaded from core.
             * Note that size and location of each label are multiplied by scaling factor in order
             * to maintain correct positions of runtime added controls on systems with differently
             * set DPI.
             */
            coreViewList = new List<System.Windows.Controls.Label>();
            for (int i = 1; i <= coreViewCount; i++)
            {
                Label label = new Label();
                label.Width = 188;
                label.Height = 16;
                label.Padding = new Thickness(2);
                label.Background = new SolidColorBrush(coreCellColor);
                label.Foreground = new SolidColorBrush(coreViewEmptyColor);
                label.FontFamily = new FontFamily("Courier New");
                label.Content = "";
                label.FontSize = 12;
                coreViewList.Add(label);
                stackPanelCoreView.Children.Add(label);
            }
            
            InitControls();
            EnableControls();
            Engine.EngineInterface = this; // Window implements interface of engine.
        }

        /* Determines core container dimensions according to engine predefined variables.
         * Builds up core container and saves it as bitmap. Resets core view box.
         */
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            InitGraphics();
        }

        public void InitGraphics()
        {
            coreColors = new List<Color>();

            // Reset core view box.
            for (int i = 0; i < coreViewCount; i++)
                coreViewList[i].Foreground = new SolidColorBrush(coreCellColor);
            coreViewAddress = 0;
            numericUpDownAddress.Value = 0;

            // Get current system DPI.
            Matrix m = PresentationSource.FromVisual(System.Windows.Application.Current.MainWindow).CompositionTarget.TransformToDevice;
            DPIx = m.M11;
            DPIy = m.M22;

            coreMap = new RenderTargetBitmap(coreContainerSize, coreContainerSize, 96 * DPIx, 96 * DPIy, PixelFormats.Pbgra32);
            DrawingVisual visual = new DrawingVisual();
            DrawingContext context = visual.RenderOpen();
            SolidColorBrush brush; // core cell color
            Pen pen; // core cell border color
            Rect rectangle; // core cell size

            // Best possible cell size is calculated according to engine core size.
            coreCellSize = 5;
            coreContainerSideCount = coreContainerSize / coreCellSize;
            coreCellCount = coreContainerSideCount * coreContainerSideCount;
            while (coreCellCount < engine.CoreSize)
            {
                if (coreCellSize > 1)
                {
                    coreCellSize--;
                    coreContainerSideCount = coreContainerSize / coreCellSize;
                    coreCellCount = coreContainerSideCount * coreContainerSideCount;
                }
            }

            // Fill bitmap with background color.
            brush = new SolidColorBrush(coreContainerColor);
            pen = new Pen(new SolidColorBrush(coreContainerColor), 0);
            rectangle = new Rect(new Size(coreContainerSize, coreContainerSize));
            context.DrawRectangle(brush, pen, rectangle);

            // Fill bitmap with core cells.
            int x = 0;
            int y = -coreCellSize;
            for (int i = 0; i < coreCellCount; i++)
            {
                if (i % coreContainerSideCount == 0) // end of line
                {
                    x = 0;  // carriage return
                    y += coreCellSize; // new line
                }
                else
                    x += coreCellSize; // next cell
                if (i < engine.CoreSize)
                    brush = new SolidColorBrush(coreCellColor); // still within core
                else
                    brush = new SolidColorBrush(coreContainerColor); // outside of core - invisible cells
                // Paint cell on bitmap.
                if (coreCellSize > 1)
                    pen = new Pen(new SolidColorBrush(coreContainerColor), penWidth);
                else
                    pen = new Pen(new SolidColorBrush(coreContainerColor), 0);
                coreColors.Add(brush.Color);
                rectangle = new Rect(new Point(x, y), new Size(coreCellSize, coreCellSize));
                
                // Detection of core view box address range.
                if (i >= coreViewAddress && i < coreViewAddress + coreViewCount && i < engine.CoreSize)
                {
                    RefreshCoreViewAt(i); // Given instruction label is requested to update itself.
                    if (checkBoxCoreView.IsChecked == true) // Slider is active. Cell must be overlayed by its color.
                    {
                        brush = new SolidColorBrush(coreViewColor);
                    }
                }

                context.DrawRectangle(brush, pen, rectangle);
            }

            context.Close();
            coreMap.Render(visual);
            imageCoreMap.Source = coreMap;
        }

        public void RestoreGraphics()
        {
            DrawingVisual visual = new DrawingVisual();
            DrawingContext context = visual.RenderOpen();
            SolidColorBrush brush; // core cell color
            Pen pen; // core cell border color
            Rect rectangle; // core cell size

            // Fill bitmap with background color.
            brush = new SolidColorBrush(coreContainerColor);
            pen = new Pen(new SolidColorBrush(coreContainerColor), 0);
            rectangle = new Rect(new Size(coreContainerSize, coreContainerSize));
            context.DrawRectangle(brush, pen, rectangle);

            // Fill bitmap with core cells.
            int x = 0;
            int y = -coreCellSize;
            for (int i = 0; i < coreCellCount; i++)
            {
                if (i % coreContainerSideCount == 0) // end of line
                {
                    x = 0;  // carriage return
                    y += coreCellSize; // new line
                }
                else
                    x += coreCellSize; // next cell
                if (i < engine.CoreSize)
                    brush = new SolidColorBrush(coreColors[i]); // still within core
                else
                    brush = new SolidColorBrush(coreContainerColor); // outside of core - invisible cells
                // Paint cell on bitmap.
                if (coreCellSize > 1)
                    pen = new Pen(new SolidColorBrush(coreContainerColor), penWidth);
                else
                    pen = new Pen(new SolidColorBrush(coreContainerColor), 0);
                rectangle = new Rect(new Point(x, y), new Size(coreCellSize, coreCellSize));

                // Detection of core view box address range.
                if (i >= coreViewAddress && i < coreViewAddress + coreViewCount && i < engine.CoreSize)
                {
                    RefreshCoreViewAt(i); // Given instruction label is requested to update itself.
                    if (checkBoxCoreView.IsChecked == true) // Slider is active. Cell must be overlayed by its color.
                    {
                        brush = new SolidColorBrush(coreViewColor);
                    }
                }

                context.DrawRectangle(brush, pen, rectangle);
            }

            context.Close();
            coreMap.Render(visual);
        }

        // Resets most of the controls in order to clear interface for another battle.
        public void InitControls()
        {
            // All warriors are wiped.
            buttonList = new List<Button>();
            progressBarList = new List<ProgressBar>();
            labelList = new List<Label>();
            sourceList = new List<string>();
            loadList = new List<string>();

            // Reset core view box.
            numericUpDownAddress.Minimum = 0;
            if (engine.CoreSize - coreViewCount < 0)
                numericUpDownAddress.Maximum = 0; // Core view box is longer than coresize. Address change is forbidden.
            else
                numericUpDownAddress.Maximum = engine.CoreSize - coreViewCount;
            numericUpDownAddress.Value = 0; // beginning of core

            // Reset battle progress bar to zero.
            progressBarBattle.Maximum = engine.MaxCycles;
            progressBarBattle.Value = 0;
            labelCycles.Content = "0 / " + progressBarBattle.Maximum.ToString();
        }

        // Removes form controls of every warrior.
        private void RemoveControls()
        {
            stackPanelWarriors.Children.Clear();
        }

        // Discovers current engine state and makes accessible appropriate controls.
        private void EnableControls()
        {
            // primitives
            bool started = engine.IsBusy;
            bool loaded = engine.QueuesCount > 0;

            // Primitives used to enable controls.
            buttonStart.IsEnabled = !started && loaded;
            buttonStop.IsEnabled = started;
            buttonStep.IsEnabled = !started && loaded;
            buttonAdd.IsEnabled = !started;
            buttonSetup.IsEnabled = !started;
            buttonWipe.IsEnabled = !started;
            checkBoxCoreView.IsEnabled = !started;

            if (started)
                labelStatus.Content = "Running...";
            else
                labelStatus.Content = "Idle";
        }

        /* Creates object array from given arguments in order to provide these arguments
         * to delegates and handlers called by Invoke function. 
         */
        private object[] ArgsToObjectArray(object object1, object object2, object object3)
        {
            object[] args = new object[3];
            args[0] = object1;
            args[1] = object2;
            args[2] = object3;
            return args;
        }

        /* NOTE: All functions which implements engine interface are called by engine
         *       itself during runtime. These functions are expected to change GUI in
         *       some way. However GUI is running on main thread while engine is running
         *       on separate thread. To maintain thread safety it is needed to invoke 
         *       GUI thread to change itself. This is achieved by Invoke function and
         *       delegate pointing to handler.
         */

        // GUI is requested to show error message.
        public void ShowError(Engine.ErrorState error)
        {
            MessageBox.Show(error.ToString("G"), "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
        }
        public delegate void ErrorCallback(Engine.ErrorState error);
        public void Error(Engine.ErrorState error)
        {
            ErrorCallback errorCallback = new ErrorCallback(this.ShowError);
            this.Dispatcher.Invoke(errorCallback, ArgsToObjectArray(error, null, null));
        }

        // GUI is requested to update task queue load of given warrior.
        public void TaskUpdate(byte warrior, bool increment, object reserved)
        {
            if (warrior < progressBarList.Count)
            {
                if (increment == true)
                    progressBarList[warrior].Value++;
                else
                    progressBarList[warrior].Value--;
                labelList[warrior].Content = progressBarList[warrior].Value.ToString() + " / " + progressBarList[warrior].Maximum.ToString();
            }
        }
        public delegate void TaskCallback(byte warrior, bool increment, object reserved);
        public void TaskChange(byte warrior, bool increment)
        {
            TaskCallback taskCallback = new TaskCallback(this.TaskUpdate);
            this.Dispatcher.Invoke(taskCallback, ArgsToObjectArray(warrior, increment, null));
        }

        // Generic delegate for void functions without arguments.
        public delegate void NoArgsCallback();

        // GUI is requested to update battle progress.
        public void ProgressUpdate()
        {
            progressBarBattle.Value++;
            labelCycles.Content = progressBarBattle.Value.ToString() + " / " + progressBarBattle.Maximum.ToString();
        }
        public void Cycle()
        {
            NoArgsCallback progressCallback = new NoArgsCallback(this.ProgressUpdate);
            progressBarBattle.Dispatcher.Invoke(progressCallback);
        }

        // GUI is requested to enable/disable controls according to engine state.
        public void StatusUpdate()
        {
            EnableControls();
        }
        public void Finished()
        {
            NoArgsCallback statusCallback = new NoArgsCallback(this.StatusUpdate);
            this.Dispatcher.Invoke(statusCallback);
        }
        public void Started()
        {
            NoArgsCallback statusCallback = new NoArgsCallback(this.StatusUpdate);
            this.Dispatcher.Invoke(statusCallback);
        }

        // GUI is requested to repaint given address in core container by given color.
        public void RepaintUpdate(Address address, Color color, object reserved)
        {
            // Address is converted to coordinates.
            int x = (address % coreContainerSideCount) * coreCellSize;
            int y = (address / coreContainerSideCount) * coreCellSize;

            // Address and color information is translated into brush and rectangle instances.
            SolidColorBrush brush = new SolidColorBrush(color);
            Rect rectangle = new Rect(new Point(x, y), new Size(coreCellSize, coreCellSize));

            DrawingVisual visual = new DrawingVisual();
            DrawingContext context = visual.RenderOpen();

            Pen pen;
            if (coreCellSize > 1)
                pen = new Pen(new SolidColorBrush(coreContainerColor), penWidth); // Cell is rectangle with 1 pixel borders.
            else
                pen = new Pen(new SolidColorBrush(coreContainerColor), 0); // Cell is 1 pixel without borders.

            context.DrawRectangle(brush, pen, rectangle);
            if (address >= coreViewAddress && address < coreViewAddress + coreViewCount)
            {
                // Cell is within address range of core view box.
                RefreshCoreViewAt(address % engine.CoreSize); // Given instruction label is requested to update itself.
                if (checkBoxCoreView.IsChecked == true && checkBoxCoreView.IsEnabled == true) // Slider is active. Cell must be overlayed by its color.
                {
                    brush = new SolidColorBrush(coreViewColor);
                    context.DrawRectangle(brush, pen, rectangle);
                }
            }

            context.Close();
            coreMap.Render(visual);
        }
        public delegate void RepaintCallback(Address address, Color color, object reserved);
        public void Repaint(Address address, Color color)
        {
            coreColors[address] = color;
            if (rendering)
            {
                RepaintCallback repaintCallback = new RepaintCallback(this.RepaintUpdate);
                try
                {
                    this.Dispatcher.Invoke(repaintCallback, ArgsToObjectArray(address, color, null));
                }
                catch (ObjectDisposedException)
                {
                    ; // Exception can occur when user closes the form even if battle is still running.
                }
            }
        }

        // Engine thread must be stopped before main window closes itself.
        private void Window_Closing(object sender, CancelEventArgs e)
        {
            engine.backgroundWorker.CancelAsync();
        }

        // User is requesting engine to start execution.
        private void buttonStart_Click(object sender, RoutedEventArgs e)
        {
            engine.step = false; // Step mechanism is disabled.
            engine.backgroundWorker.RunWorkerAsync();
        }

        // User is requesting engine to stop execution.
        private void buttonStop_Click(object sender, RoutedEventArgs e)
        {
            engine.backgroundWorker.CancelAsync();
        }

        // User is changing execution speed of engine.
        private void sliderSpeed_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            engine.SleepLimit = Convert.ToInt32(sliderSpeed.Value * 5);
        }

        // User is requesting engine to execute one instruction of every living warrior.
        private void buttonStep_Click(object sender, RoutedEventArgs e)
        {
            engine.step = true; // Step mechanism is enabled.
            engine.backgroundWorker.RunWorkerAsync();
        }

        /* Each warrior has its own button in GUI. This function is handler for
         * all these buttons. When user clicks on button, window with warrior's
         * assembly file and load file appears on the screen. Index of warrior
         * is derived from its button text label.
         */
        private void buttonColored_Click(object sender, RoutedEventArgs e)
        {
            windowWarrior = new WindowWarrior();
            windowWarrior.textBoxSource.Text = sourceList[Int32.Parse((sender as Button).Content.ToString()) - 1];
            windowWarrior.textBoxLoad.Text = loadList[Int32.Parse((sender as Button).Content.ToString()) - 1];
            windowWarrior.Title = "Warrior " + (sender as Button).Content.ToString();
            windowWarrior.Show();
        }

        /* When user hoover mouse over warrior's button, warrior is requested to invoke
         * window to highlight warrior's current instruction by different color than other
         * warriors. Index of warrior is derived from its button text label.
         */
        private void buttonColored_MouseEnter(object sender, MouseEventArgs e)
        {
            Engine.Warrior warrior;
            if (engine.QueuesAtID(Int32.Parse((sender as Button).Content.ToString()) - 1, out warrior) == true)
            {
                warrior.selected = true;
                warrior.PaintExecute();
            }
        }
        private void buttonColored_MouseLeave(object sender, MouseEventArgs e)
        {
            Engine.Warrior warrior;
            if (engine.QueuesAtID(Int32.Parse((sender as Button).Content.ToString()) - 1, out warrior) == true)
            {
                warrior.selected = false;
                warrior.PaintExecute();
            }
        }

        /* Open file dialog is shown. User chooses redcode warrior file. File is parsed and
         * handed over to engine loader function. If any error occurs, user is notified by
         * message box. New warrior's controls are added to main window.
         */
        private void buttonAdd_Click(object sender, RoutedEventArgs e)
        {
            if (buttonList.Count < engine.Warriors) // Checks whether maximum warriors count isn't reached.
            {
                System.Windows.Forms.OpenFileDialog openFileDialog1 = new System.Windows.Forms.OpenFileDialog();
                openFileDialog1.DefaultExt = "red";
                openFileDialog1.Filter = "Redcode file|*.red";

                System.Windows.Forms.DialogResult dlgResult;
                dlgResult = openFileDialog1.ShowDialog();
                if (dlgResult == System.Windows.Forms.DialogResult.OK) // User selected file.
                {
                    StreamReader reader;
                    reader = new StreamReader(openFileDialog1.FileName);
                    // Redcode file must starts with ;redcode comment on the first line.
                    string identifier = ";redcode";
                    for (int i = 0; i < 8; i++)
                    {
                        if (identifier[i] != (char)reader.Peek())
                        {
                            MessageBox.Show("Not a redcode file", "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
                            return;
                        }
                        reader.Read();
                    }
                    // New instance of parser is created. File stream is parsed and any exception is catched and reported.
                    reader = new StreamReader(openFileDialog1.FileName);
                    Parser parser = new Parser(this.engine);
                    Stream parsed = new MemoryStream();
                    try
                    {
                        parsed = parser.Parse(reader);
                    }
                    catch (ApplicationException appException)
                    {
                        MessageBox.Show(appException.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
                        return;
                    }
                    catch (DivideByZeroException divException)
                    {
                        MessageBox.Show(divException.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
                        return;
                    }
                    catch (Exception exception)
                    {
                        MessageBox.Show("Parser reported error. File is not valid redcode assembly file. (" + exception.Message + ")", "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
                        return;
                    }
                    // Load file is handed over to engine loader. Errors are reported to user.
                    Engine.ErrorState error = engine.Loader(parsed);
                    if (error != Engine.ErrorState.Normal)
                    {
                        MessageBox.Show("Load error: " + error.ToString("G"), "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
                        return;
                    }
                    // Assembly file and load file of warrior are saved for later use.
                    reader = new StreamReader(openFileDialog1.FileName);
                    sourceList.Add(reader.ReadToEnd());
                    parsed.Seek(0, SeekOrigin.Begin);
                    reader = new StreamReader(parsed);
                    loadList.Add(reader.ReadToEnd());

                    /* Each warrior is represented in GUI by its source code button and task queue 
                     * progress bar with label. Note that size and location of each control are multiplied 
                     * by scaling factor in order to maintain correct positions of runtime added controls 
                     * on systems with differently set DPI.
                     */
                    StackPanel panel = new StackPanel();
                    panel.Orientation = Orientation.Horizontal;
                    StackPanel subPanel = new StackPanel();
                    subPanel.Orientation = Orientation.Vertical;
                    Button button = new Button();
                    ProgressBar progressBar = new ProgressBar();
                    Label label = new Label();
                    button.Width = 39;
                    button.Height = 23;
                    progressBar.Width = 124;
                    progressBar.Height = 10;
                    label.Width = 123;
                    label.Height = 14;

                    subPanel.Children.Add(label);
                    subPanel.Children.Add(progressBar);
                    panel.Children.Add(button);
                    panel.Children.Add(subPanel);

                    // Controls are initialized.
                    button.Content = (buttonList.Count + 1).ToString();
                    button.Background = new SolidColorBrush(Engine.colors[buttonList.Count % Engine.colors.Length]); // Button is colored by warrior's color.
                    button.Click += new RoutedEventHandler(this.buttonColored_Click);
                    button.MouseEnter += new MouseEventHandler(this.buttonColored_MouseEnter);
                    button.MouseLeave += new MouseEventHandler(this.buttonColored_MouseLeave);
                    progressBar.Value = 1; // Before execution starts, each warrior has one active task in task queue.
                    progressBar.Maximum = engine.MaxProcesses;
                    label.FontFamily = new FontFamily("Microsoft Sans Serif");
                    label.Padding = new Thickness(0);
                    label.FontSize = 8;
                    label.Content = "1 / " + engine.MaxProcesses.ToString();

                    // Controls are added to main window and saved for later use.
                    buttonList.Add(button);
                    progressBarList.Add(progressBar);
                    labelList.Add(label);
                    stackPanelWarriors.Children.Add(panel);

                    EnableControls();
                }
            }
            else
                MessageBox.Show("Maximum warrior count reached", "Error", MessageBoxButton.OK, MessageBoxImage.Stop);
        }

        /* When user clicks on Setup button, new setup window is created and 
         * engine current variable set is loaded into window's controls. After
         * variables are loaded, window is shown as a dialog. If user commits
         * changes, all values are offered to engine. Engine properties are
         * internally controlled and if it is possible, new value is set.
         * Setup window then compare offered values with values that engine
         * accepted. If there is any difference, user is reported about
         * adjustments. Core is then rebuilt, graphics is reset and GUI is
         * cleared.
         */
        private void buttonSetup_Click(object sender, RoutedEventArgs e)
        {
            WindowSetup windowSetup = new WindowSetup();

            // Variables are loaded from engine into setup window controls.
            windowSetup.numericUpDownCoreSize.Minimum = 1;
            windowSetup.numericUpDownCoreSize.Maximum = 230400; // 480 * 480
            windowSetup.numericUpDownCoreSize.Value = engine.CoreSize;

            windowSetup.numericUpDownMaxCycles.Minimum = 1;
            windowSetup.numericUpDownMaxCycles.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownMaxCycles.Value = engine.MaxCycles;

            windowSetup.numericUpDownMaxLength.Minimum = 1;
            windowSetup.numericUpDownMaxLength.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownMaxLength.Value = engine.MaxLength;

            windowSetup.numericUpDownMaxProcesses.Minimum = 1;
            windowSetup.numericUpDownMaxProcesses.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownMaxProcesses.Value = engine.MaxProcesses;

            windowSetup.numericUpDownMinDistance.Minimum = 0;
            windowSetup.numericUpDownMinDistance.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownMinDistance.Value = engine.MinDistance;

            windowSetup.numericUpDownReadDistance.Minimum = 1;
            windowSetup.numericUpDownReadDistance.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownReadDistance.Value = engine.ReadDistance;
            if (engine.ReadDistance == engine.CoreSize)
            {
                windowSetup.numericUpDownReadDistance.Enabled = false;
                windowSetup.checkBoxRead.IsChecked = true;
            }
            else
                windowSetup.checkBoxRead.IsChecked = false;

            windowSetup.numericUpDownWriteDistance.Minimum = 1;
            windowSetup.numericUpDownWriteDistance.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownWriteDistance.Value = engine.WriteDistance;
            if (engine.WriteDistance == engine.CoreSize)
            {
                windowSetup.numericUpDownWriteDistance.Enabled = false;
                windowSetup.checkBoxWrite.IsChecked = true;
            }
            else
                windowSetup.checkBoxWrite.IsChecked = false;

            windowSetup.numericUpDownWarriors.Minimum = 2;
            windowSetup.numericUpDownWarriors.Maximum = Byte.MaxValue;
            windowSetup.numericUpDownWarriors.Value = engine.Warriors;

            windowSetup.numericUpDownSeparation.Minimum = 0;
            windowSetup.numericUpDownSeparation.Maximum = Int32.MaxValue;
            if (engine.SeparationModifier == 1)
            {
                windowSetup.numericUpDownSeparation.Value = engine.Separation;
                windowSetup.checkBoxSeparation.IsChecked = false;
            }
            else
            {
                windowSetup.numericUpDownSeparation.Value = 0;
                windowSetup.numericUpDownSeparation.Enabled = false;
                windowSetup.checkBoxSeparation.IsChecked = true;
            }

            for (int i = 0; i < Engine.opcode.Length; i++)
            {
                windowSetup.comboBoxOpcode.Items.Add(Engine.opcode[i].ToString("G"));
                if (engine.InitialInstruction.opcode == Engine.opcode[i])
                    windowSetup.comboBoxOpcode.SelectedItem = windowSetup.comboBoxOpcode.Items[i];
            }
            for (int i = 0; i < Engine.modifier.Length; i++)
            {
                windowSetup.comboBoxModifier.Items.Add(Engine.modifier[i].ToString("G"));
                if (engine.InitialInstruction.modifier == Engine.modifier[i])
                    windowSetup.comboBoxModifier.SelectedItem = windowSetup.comboBoxModifier.Items[i];
            }
            for (int i = 0; i < Engine.mode.Length; i++)
            {
                windowSetup.comboBoxAMode.Items.Add(Parser.modes[i].ToString());
                windowSetup.comboBoxBMode.Items.Add(Parser.modes[i].ToString());
                if (engine.InitialInstruction.aMode == Engine.mode[i])
                    windowSetup.comboBoxAMode.SelectedItem = windowSetup.comboBoxAMode.Items[i];
                if (engine.InitialInstruction.bMode == Engine.mode[i])
                    windowSetup.comboBoxBMode.SelectedItem = windowSetup.comboBoxBMode.Items[i];
            }
            windowSetup.numericUpDownANumber.Minimum = 0;
            windowSetup.numericUpDownANumber.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownANumber.Value = engine.InitialInstruction.aNumber;
            windowSetup.numericUpDownBNumber.Minimum = 0;
            windowSetup.numericUpDownBNumber.Maximum = Int32.MaxValue;
            windowSetup.numericUpDownBNumber.Value = engine.InitialInstruction.bNumber;
            if (engine.InitialInstructionModifier == 2)
            {
                windowSetup.comboBoxInitialInstruction.SelectedItem = windowSetup.comboBoxInitialInstruction.Items[2];
                windowSetup.comboBoxOpcode.IsEnabled = false;
                windowSetup.comboBoxModifier.IsEnabled = false;
                windowSetup.comboBoxAMode.IsEnabled = false;
                windowSetup.numericUpDownANumber.Enabled = false;
                windowSetup.comboBoxBMode.IsEnabled = false;
                windowSetup.numericUpDownBNumber.Enabled = false;
            }
            else if (engine.InitialInstructionModifier == 1)
            {
                windowSetup.comboBoxInitialInstruction.SelectedItem = windowSetup.comboBoxInitialInstruction.Items[1];
                windowSetup.comboBoxOpcode.IsEnabled = false;
                windowSetup.comboBoxModifier.IsEnabled = false;
                windowSetup.comboBoxAMode.IsEnabled = false;
                windowSetup.numericUpDownANumber.Enabled = false;
                windowSetup.comboBoxBMode.IsEnabled = false;
                windowSetup.numericUpDownBNumber.Enabled = false;
            }
            else
            {
                windowSetup.comboBoxInitialInstruction.SelectedItem = windowSetup.comboBoxInitialInstruction.Items[0];
                windowSetup.comboBoxOpcode.IsEnabled = true;
                windowSetup.comboBoxModifier.IsEnabled = true;
                windowSetup.comboBoxAMode.IsEnabled = true;
                windowSetup.numericUpDownANumber.Enabled = true;
                windowSetup.comboBoxBMode.IsEnabled = true;
                windowSetup.numericUpDownBNumber.Enabled = true;
            }
            if (Parser.conversion88 == true)
                windowSetup.comboBoxConversion.SelectedItem = windowSetup.comboBoxConversion.Items[0];
            else
                windowSetup.comboBoxConversion.SelectedItem = windowSetup.comboBoxConversion.Items[1];

            // Window is prepared and can be showed to user as a dialog.
            int result = windowSetup.DoModal();

            // Changes are offered to engine. Refused values are adjusted and reported to user.
            if (result == WindowSetup.Ok)
            {
                engine.CoreSize = Convert.ToInt32(windowSetup.numericUpDownCoreSize.Value);
                if (engine.CoreSize != windowSetup.numericUpDownCoreSize.Value)
                    MessageBox.Show("Core size " + windowSetup.numericUpDownCoreSize.Value.ToString() + " is not possible to set. Using value " + engine.CoreSize + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.MaxCycles = Convert.ToInt32(windowSetup.numericUpDownMaxCycles.Value);
                if (engine.MaxCycles != windowSetup.numericUpDownMaxCycles.Value)
                    MessageBox.Show("Max cycles " + windowSetup.numericUpDownMaxCycles.Value.ToString() + " is not possible to set. Using value " + engine.MaxCycles + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.MaxProcesses = Convert.ToInt32(windowSetup.numericUpDownMaxProcesses.Value);
                if (engine.MaxProcesses != windowSetup.numericUpDownMaxProcesses.Value)
                    MessageBox.Show("Max processes " + windowSetup.numericUpDownMaxProcesses.Value.ToString() + " is not possible to set. Using value " + engine.MaxProcesses + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.MaxLength = Convert.ToInt32(windowSetup.numericUpDownMaxLength.Value);
                if (engine.MaxLength != windowSetup.numericUpDownMaxLength.Value)
                    MessageBox.Show("Max length " + windowSetup.numericUpDownMaxLength.Value.ToString() + " is not possible to set. Using value " + engine.MaxLength + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.MinDistance = Convert.ToInt32(windowSetup.numericUpDownMinDistance.Value);
                if (engine.MinDistance != windowSetup.numericUpDownMinDistance.Value)
                    MessageBox.Show("Min distance " + windowSetup.numericUpDownMinDistance.Value.ToString() + " is not possible to set. Using value " + engine.MinDistance + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.ReadDistance = Convert.ToInt32(windowSetup.numericUpDownReadDistance.Value);
                if (engine.ReadDistance != windowSetup.numericUpDownReadDistance.Value)
                    MessageBox.Show("Read distance " + windowSetup.numericUpDownReadDistance.Value.ToString() + " is not possible to set. Using value " + engine.ReadDistance + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.WriteDistance = Convert.ToInt32(windowSetup.numericUpDownWriteDistance.Value);
                if (engine.WriteDistance != windowSetup.numericUpDownWriteDistance.Value)
                    MessageBox.Show("Write distance " + windowSetup.numericUpDownWriteDistance.Value.ToString() + " is not possible to set. Using value " + engine.WriteDistance + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                engine.Separation = Convert.ToInt32(windowSetup.numericUpDownSeparation.Value);
                if (engine.Separation != windowSetup.numericUpDownSeparation.Value)
                    MessageBox.Show("Separation " + windowSetup.numericUpDownSeparation.Value.ToString() + " is not possible to set. Using value " + engine.Separation + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                if (windowSetup.checkBoxSeparation.IsChecked == true)
                    engine.SeparationModifier = 0;
                else
                    engine.SeparationModifier = 1;
                engine.Warriors = Convert.ToByte(windowSetup.numericUpDownWarriors.Value);
                if (engine.Warriors != windowSetup.numericUpDownWarriors.Value)
                    MessageBox.Show("Warriors " + windowSetup.numericUpDownWarriors.Value.ToString() + " is not possible to set. Using value " + engine.Warriors + ".", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                if (windowSetup.comboBoxInitialInstruction.SelectedIndex == 0)
                {
                    engine.InitialInstructionModifier = 0;
                    Engine.Instruction instruction = new Engine.Instruction();
                    for (int i = 0; i < Engine.opcode.Length; i++)
                    {
                        if (Engine.opcode[i].ToString("G") == windowSetup.comboBoxOpcode.SelectedItem.ToString())
                        {
                            instruction.opcode = Engine.opcode[i];
                            break;
                        }
                    }
                    for (int i = 0; i < Engine.modifier.Length; i++)
                    {
                        if (Engine.modifier[i].ToString("G") == windowSetup.comboBoxModifier.SelectedItem.ToString())
                        {
                            instruction.modifier = Engine.modifier[i];
                            break;
                        }
                    }
                    for (int i = 0; i < Engine.mode.Length; i++)
                    {
                        if (Parser.modes[i].ToString() == windowSetup.comboBoxAMode.SelectedItem.ToString())
                        {
                            instruction.aMode = Engine.mode[i];
                        }
                        if (Parser.modes[i].ToString() == windowSetup.comboBoxBMode.SelectedItem.ToString())
                        {
                            instruction.bMode = Engine.mode[i];
                        }
                    }
                    instruction.aNumber = Convert.ToInt32(windowSetup.numericUpDownANumber.Value);
                    instruction.bNumber = Convert.ToInt32(windowSetup.numericUpDownBNumber.Value);
                    engine.InitialInstruction = instruction;
                    if (engine.InitialInstruction.aNumber != instruction.aNumber || engine.InitialInstruction.bNumber != instruction.bNumber)
                        MessageBox.Show("Initial instruction operands are out of range. Using closest possible values.", "Notification", MessageBoxButton.OK, MessageBoxImage.Information);
                }
                else
                    engine.InitialInstructionModifier = Convert.ToByte(windowSetup.comboBoxInitialInstruction.SelectedIndex);
                if (windowSetup.comboBoxConversion.SelectedIndex == 0)
                    Parser.conversion88 = true;
                else
                    Parser.conversion88 = false;

                // Data structures in engine and GUI are wiped.
                engine.RebuildCore();
                InitGraphics();
                RemoveControls();
                InitControls();
                EnableControls();
            }
        }

        // Shortcut to OK button in Setup window. Resets data structures without changing current variable set.
        private void buttonWipe_Click(object sender, RoutedEventArgs e)
        {
            engine.RebuildCore();
            InitGraphics();
            RemoveControls();
            InitControls();
            EnableControls();
        }

        // Shows about box as a dialog.
        private void buttonAbout_Click(object sender, RoutedEventArgs e)
        {
            WindowAbout about = new WindowAbout();
            about.DoModal();
        }

        // Enables core view box slider. Affected cells are invoked to repaint.
        private void checkBoxCoreView_CheckedChanged(object sender, RoutedEventArgs e)
        {
            int i = 0;
            for (Address address = coreViewAddress; address < coreViewAddress + coreViewCount && address < engine.CoreSize; address++)
            {
                if ((coreViewList[i].Foreground as SolidColorBrush).Color == coreViewEmptyColor)
                    Repaint(address, coreCellColor);
                else
                    Repaint(address, (coreViewList[i].Foreground as SolidColorBrush).Color);
                i++;
            }
        }
        private void checkBoxCoreView_EnabledChanged(object sender, RoutedEventArgs e)
        {
            checkBoxCoreView_CheckedChanged(sender, e);
        }

        private void checkBoxCoreMap_CheckedChanged(object sender, RoutedEventArgs e)
        {
            if (checkBoxCoreMap.IsChecked == true)
                rendering = true;
            else
                rendering = false;
            if (checkBoxCoreMap.IsChecked == true && coreColors != null)
                RestoreGraphics();
        }

        // Loads instructions from given address into core view box.
        private void LoadCoreView(Address newAddress)
        {
            Address oldAddress = coreViewAddress; // Old address range is temporarily remembered.
            coreViewAddress = newAddress; // New address range is updated.
            /* In case that slider is active and thus cells must be repainted,
             * it must be decided whether cells are invoked from left to right or
             * from right to left side of the slider.
             */
            if (oldAddress < newAddress) // updated from left to right
            {
                for (int i = 0; i < coreViewCount; i++) // Whole core view box must be updated.
                {
                    if (checkBoxCoreView.IsChecked == true && checkBoxCoreView.IsEnabled == true)
                    {
                        /* Slider cells on old position are successively replaced by their original colors
                         * which are stored in corresponding labels as a font colors.
                         */
                        if ((coreViewList[i].Foreground as SolidColorBrush).Color == coreViewEmptyColor)
                            Repaint((oldAddress + i) % engine.CoreSize, coreCellColor); // Cell was empty.
                        else
                            Repaint((oldAddress + i) % engine.CoreSize, (coreViewList[i].Foreground as SolidColorBrush).Color); // Cell was influenced by warrior.
                    }
                    // Instruction label is updated with new instruction string corresponding to new address range.
                    RefreshCoreViewAt((coreViewAddress + i) % engine.CoreSize);
                    if (checkBoxCoreView.IsChecked == true && checkBoxCoreView.IsEnabled == true)
                    {
                        // If slider is active, it is painted cell after cell onto new position.
                        if ((coreViewList[i].Foreground as SolidColorBrush).Color == coreViewEmptyColor)
                            Repaint((newAddress + i) % engine.CoreSize, coreCellColor);
                        else
                            Repaint((newAddress + i) % engine.CoreSize, (coreViewList[i].Foreground as SolidColorBrush).Color);
                    }
                }
            }
            else if (oldAddress > newAddress) // updated from right to left
            {
                for (int i = coreViewCount - 1; i >= 0; i--)
                {
                    if (checkBoxCoreView.IsChecked == true && checkBoxCoreView.IsEnabled == true)
                    {
                        if ((coreViewList[i].Foreground as SolidColorBrush).Color == coreViewEmptyColor)
                            Repaint((oldAddress + i) % engine.CoreSize, coreCellColor);
                        else
                            Repaint((oldAddress + i) % engine.CoreSize, (coreViewList[i].Foreground as SolidColorBrush).Color);
                    }
                    RefreshCoreViewAt((coreViewAddress + i) % engine.CoreSize);
                    if (checkBoxCoreView.IsChecked == true && checkBoxCoreView.IsEnabled == true)
                    {
                        if ((coreViewList[i].Foreground as SolidColorBrush).Color == coreViewEmptyColor)
                            Repaint((newAddress + i) % engine.CoreSize, coreCellColor);
                        else
                            Repaint((newAddress + i) % engine.CoreSize, (coreViewList[i].Foreground as SolidColorBrush).Color);
                    }
                }
            }
        }

        /* Label in core view box is updated by instruction which is located on given address
        * in core. Index of label is derived from given address and core view box initial
        * address.
        */
        private void RefreshCoreViewAt(Address address)
        {
            string output; // New instruction string of label.
            Color color; // New color of label.

            /* Instruction is loaded from core and converted into readable string.
             * Note that instruction is aligned into 3 columns.
             */
            Engine.Instruction instruction = engine.CoreAt(address);
            // Opcode and modifier is added to string.
            output = instruction.opcode.ToString("G") + "." + instruction.modifier.ToString("G");
            // Space is added to string.
            if (output.Length == 5)
                output += "  "; // Modifier was only 1 character.
            else
                output += " "; // Modifier was composed of 2 characters.
            // A-mode index is identified.
            int i = 0;
            for (i = 0; i < Engine.mode.Length; i++)
                if (Engine.mode[i] == instruction.aMode)
                    break;
            // A-mode and A-number are joined. Note that mode index is used to access mode array in parser.
            string partA = Parser.modes[i].ToString() + instruction.aNumber.ToString() + ",";
            // B-mode index is identified.
            for (i = 0; i < Engine.mode.Length; i++)
                if (Engine.mode[i] == instruction.bMode)
                    break;
            // B-mode and B-number are joined. Note that mode index is used to access mode array in parser.
            string partB = Parser.modes[i].ToString() + instruction.bNumber.ToString();
            /* Both operands are added to output string and aligned by spaces into columns.
             * Note that max core size in this implementation is 230400. Along with sign and
             * mode character, final string might be at most 8 characters long. A-operand is
             * in addition terminated by comma.
             */
            for (i = 0; i < 9 - partA.Length; i++)
                output += " ";
            output += partA;
            for (i = 0; i < 8 - partB.Length; i++)
                output += " ";
            output += partB;
            color = coreColors[address];
            /* If cell is empty, color must be changed because in current settings,
             * both empty cell and background of core view box are the same.
             */
            if (color.A == coreCellColor.A && color.B == coreCellColor.B && color.G == coreCellColor.G && color.R == coreCellColor.R)
                color = coreViewEmptyColor;
            // Index of altered label is calculated.
            int index = address - coreViewAddress;
            // Label is updated.
            coreViewList[index].Content = output;
            coreViewList[index].Foreground = new SolidColorBrush(color);
        }

        // Address range of core view box can be changed directly.
        private void numericUpDownAddress_ValueChanged(object sender, EventArgs e)
        {
            LoadCoreView(Convert.ToInt32((sender as System.Windows.Forms.NumericUpDown).Value));
        }

        /* Address range of core view box can be changed indirectly through 
         * mouse click into core container.
         */
        private void Window_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Point pos = e.GetPosition(imageCoreMap);
            int x = Convert.ToInt32(pos.X * (coreContainerSize / imageCoreMap.ActualWidth));
            int y = Convert.ToInt32(pos.Y * (coreContainerSize / imageCoreMap.ActualHeight));
            if (x < coreContainerSize && y < coreContainerSize) // Mouse click is within core container.
            {
                // Coordinates of click are translated into address.
                int address = (y / coreCellSize) * coreContainerSideCount + (x / coreCellSize + 1) - 1;
                /* Core container can be partially empty. It is checked whether address is lower than core size.
                 * Note that there must be also reserved place for core view box slider.
                 */
                if (address <= engine.CoreSize - coreViewCount)
                    numericUpDownAddress.Value = address;
            }
        }
    }
}
